#ifdef DEBUG

#include <stdarg.h>
#include <stdio.h>
#include "putty.h"
#include "wince.h"
#include "wince_device.h"

static FILE *debug_fp = NULL;

#ifdef DEBUG_UDP
#include <winsock2.h>

#pragma comment(lib , "ws2")
static SOCKET udp_socket = INVALID_SOCKET;
static unsigned short udp_port = 9998;

static int udplog_bind()
{
	int ret;
	SOCKADDR_IN addr;
	
	addr.sin_family = AF_INET;
	addr.sin_port = htons(udp_port);
	addr.sin_addr.s_addr = htonl(INADDR_ANY);
	
	ret = bind(udp_socket, (struct sockaddr *)&addr, sizeof(addr));
	
	return (ret == 0);
}

void udplog_init()
{
	int ret;
	int b = TRUE;
	
	if (udp_socket != INVALID_SOCKET)
		return;
		
	udp_socket = socket(PF_INET, SOCK_DGRAM, 0);

	if (udp_socket == INVALID_SOCKET) {
		MessageBox(0, L"Can't create socket", L"PocketPutty", MB_ICONERROR);
		goto error;
	}
	
	ret = setsockopt(udp_socket, SOL_SOCKET, SO_BROADCAST, (char*)&b, sizeof(b));
	
	if (ret != 0) {
		MessageBox(NULL, L"setsockopt failed", L"PocketPutty", MB_ICONERROR);
		goto error;
	}
	
	if (udplog_bind())
		return;
		
	MessageBox(NULL, L"Can't bind to port 9998", L"PocketPutty", MB_ICONERROR);

error:
	if (udp_socket != INVALID_SOCKET)
		closesocket(udp_socket);
}

static void udplog_send(const char *buf)
{
	SOCKADDR_IN sa;
	sa.sin_family = AF_INET;
	sa.sin_port = htons(udp_port);
	sa.sin_addr.s_addr = htonl(INADDR_BROADCAST);

	sendto(udp_socket, buf, strlen(buf), 0, (struct sockaddr*)&sa, sizeof(sa));
}

void udplog_cleanup()
{
	if (udp_socket != INVALID_SOCKET) {
		shutdown(udp_socket, SD_BOTH);
		closesocket(udp_socket);
	}
}
#endif /* ifdef DEBUG_UDP */

static void dputs_wce(const char *buf)
{
	if (!debug_fp) {
		switch (wince_device) {
		case WINCE_DEV_EVC4EMU_PC2K3:
		case WINCE_DEV_VS2K5EMU_PC2K3:
		case WINCE_DEV_VS2K5EMU_SP2K3:
		case WINCE_DEV_VS2K5EMU_WM5:
			debug_fp = fopen("\\Storage Card\\putty_debug.log", "w");
			if (!debug_fp) debug_fp = fopen("\\Storage\\putty_debug.log", "w");
			if (!debug_fp) debug_fp = fopen("\\putty_debug.log", "w");
			break;
		default:
			debug_fp = fopen("\\Ramdisk\\putty_debug.log", "w");
			if (!debug_fp) debug_fp = fopen("\\Ramdrive\\putty_debug.log", "w");
			if (!debug_fp) debug_fp = fopen("\\Storage Card\\putty_debug.log", "w");
			if (!debug_fp) debug_fp = fopen("\\Storage\\putty_debug.log", "w");
			if (!debug_fp) debug_fp = fopen("\\putty_debug.log", "w");
		}
	}
	
	fputs(buf, debug_fp);
	fflush(debug_fp);
}

void dputs(char *buf)
{
	dputs_wce(buf);
}

void debug_printf_level(int level, char *fmt, ...)
{
	char *buf;
	va_list ap;
	
	if (level & debug_level) {
		va_start(ap, fmt);
		buf = dupvprintf(fmt, ap);
#ifdef DEBUG_UDP
		udplog_send(buf);
#else
		dputs(buf);
#endif
		sfree(buf);
		va_end(ap);
	}
}

#endif /* #ifdef DEBUG */
