#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <ctype.h>
#if (UNDER_CE < 0x400) & defined(_MSC_VER)
#define assert(x)
#include "wince_time.h"
#else
#include <assert.h>
#include <time.h>
#endif

#include "putty.h"
#include "ssh.h"
#include "wince_res.h"
#include "storage.h"
#include "dialog.h"

#include <imm.h>
#include <commctrl.h>
#include <commdlg.h>
#include "wince_aygshell.h"

#ifdef USE_RELAYOUTDIALOG
#include "RelayoutDialog.h"
#endif

#ifdef __GNUC__
#include <shellapi.h>
#define _SHELLEXECUTEINFO _SHELLEXECUTEINFOW
#define StringCbPrintf _sntprintf

typedef struct DLGTEMPLATEEX
{
	WORD dlgVer;
	WORD signature;
	DWORD helpID;
	DWORD exStyle;
	DWORD style;
	WORD cDlgItems;
	SHORT x;
	SHORT y;
	SHORT cx;
	SHORT cy;
} DLGTEMPLATEEX, *LPDLGTEMPLATEEX;
#endif

/*
 * These are the various bits of data required to handle the
 * portable-dialog stuff in the config box. Having them at file
 * scope in here isn't too bad a place to put them; if we were ever
 * to need more than one config box per process we could always
 * shift them to a per-config-box structure stored in GWL_USERDATA.
 */
static struct controlbox *ctrlbox;
/*
 * ctrls_base holds the OK and Cancel buttons: the controls which
 * are present in all dialog panels. ctrls_panel holds the ones
 * which change from panel to panel.
 */
static struct winctrls ctrls_base, ctrls_panel;
static struct dlgparam dp;

static char **events = NULL;
static int nevents = 0, negsize = 0;

extern Conf *conf;				/* defined in wince_window.c */

#define STRSAFE_NO_DEPRECATE
#include <strsafe.h>

#define PRINTER_DISABLED_STRING "None (printing disabled)"

#if defined(__GNUC__) && defined(__COREDLL__)
extern int _wtoi(const wchar_t*);		/* defined in wince_misc.c */
#endif

static HMENU popup_menu;
static UINT override_menu_height;

static HWND CreateRpCommandBar(HWND hwnd);

static BOOL IsLandscape(void);

#if UNDER_CE < 0x400 && !defined(__GNUC__)
BOOL WINAPI CheckDlgButton(HWND hDlg, int nIDButton, UINT uCheck)
{
    return SendDlgItemMessage(hDlg, nIDButton, BM_SETCHECK, \
			      (WPARAM)uCheck, (LPARAM)0);
}
#endif

#define M_ISDLGBUTTONCHECKED(hDlg, nIDButton) \
	(SendMessage(GetDlgItem(hDlg, nIDButton), \
		     BM_GETCHECK, (WPARAM)0, (LPARAM)0))

#define M_CHECKDLGBUTTON(hDlg, nIDButton, uCheck) \
	CheckDlgButton(hDlg, nIDButton, uCheck)

void force_normal(HWND hwnd)
{
    static int recurse = 0;

    WINDOWPLACEMENT wp;

    if (recurse)
	return;
    recurse = 1;

    wp.length = sizeof(wp);
#if 0
    if (GetWindowPlacement(hwnd, &wp) && wp.showCmd == SW_SHOWMAXIMIZED) {
	wp.showCmd = SW_SHOWNORMAL;
	SetWindowPlacement(hwnd, &wp);
    }
#endif
    recurse = 0;
}

static INT_PTR CALLBACK LicenceProc(HWND hwnd, UINT msg,
				    WPARAM wParam, LPARAM lParam)
{
#ifdef SHELL_AYGSHELL
	static SHACTIVATEINFO sai;
#endif

	debug_level((DBG_WMSG, "\nLicenseProc Message %d(%x, %x)\n", msg, wParam, lParam));

	switch (msg) {
	case WM_INITDIALOG:
		debug_level((DBG_WMSG, "WM_INITDLG\n"));
		{
		RECT rt;
		char *str = dupprintf("%s Licence", appname);
		WCHAR *caption = ansi2unicode(str);
#ifdef SHELL_AYGSHELL
		{
		SHINITDLGINFO shidi;
		/* Create a Done button and size it. */
		shidi.dwMask = SHIDIM_FLAGS;
		shidi.dwFlags = SHIDIF_DONEBUTTON | SHIDIF_SIPDOWN | SHIDIF_SIZEDLGFULLSCREEN |
				SHIDIF_EMPTYMENU;
		shidi.hDlg = hwnd;
		if (as_SHInitDialog)
			as_SHInitDialog(&shidi);
		/* Register */
		ZeroMemory(&sai, sizeof(sai));
		sai.cbSize = sizeof(sai);
		}
#endif
		//SetWindowTextW(hwnd, caption);
		sfree(caption);
		sfree(str);
		GetClientRect(hwnd, &rt);
		MoveWindow(GetDlgItem(hwnd, IDC_E_LICENCE), rt.left, rt.top, rt.right - rt.left, rt.bottom - rt.top, FALSE);
		SetDlgItemTextW(hwnd, IDC_E_LICENCE,
			L"PuTTY is copyright 1997-2015 Simon Tatham.\r\n\r\n"
			L"Portions \r\ncopyright Robert de Bath, Joris van Rantwijk, "
			L"Delian Delchev, Andreas Schultz, Jeroen Massar, Wez Furlong, "
			L"Nicolas Barry, CORE SDI S.A. and Aleq Berka\r\n\r\n"
			L"Permission is hereby granted, free of charge, to any person "
			L"obtaining a copy of this software and associated documentation "
			L"files (the \"Software\"), to deal in the Software without restriction "
			L"including without limitation the rights to use, copy, modify, merge, "
			L"publish, distribute, sublicense, and/or sell copies of the Software, and "
			L"to permit persons to whom the Software is furnished to do so, subject to "
			L"the following conditions:\r\n\r\n"
			L"The above copyright notice and this permission notice shall be included in "
			L"all copies or substantial portions of the Software.\r\n\r\n"
			L"THE SOFTWARE IS PROVIDED \"AS IS\", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR "
			L"IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS "
			L"FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL SIMON TATHAM BE "
			L"LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, "
			L"TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE "
			L"OR OTHER DEALINGS IN THE SOFTWARE.");
		SendMessage(GetDlgItem(hwnd, IDC_E_LICENCE), EM_SETSEL, -1, 0);
		break;
		}
		return 1;
	case WM_COMMAND:
		debug_level((DBG_WMSG, "WM_COMMAND\n"));
		switch (LOWORD(wParam)) {
		case IDOK:
		case IDCANCEL:
			debug_level((DBG_WMSG, "IDOK/IDCANCEL\n"));
			EndDialog(hwnd, TRUE);
			break;
		}
		break;
	case WM_CLOSE:
		debug_level((DBG_WMSG, "WM_CLOSE\n"));
		EndDialog(hwnd, TRUE);
		break;
	case WM_ACTIVATE:
		debug_level((DBG_WMSG, "WM_ACTIVATE\n"));
#ifdef SHELL_AYGSHELL
		if (as_SHHandleWMActivate)
			as_SHHandleWMActivate(hwnd, wParam, lParam, &sai, FALSE);
#endif
		break;
	case WM_SETTINGCHANGE:
		debug_level((DBG_WMSG, "WM_SETTINGCHANGE\n"));
#ifdef SHELL_AYGSHELL
		if (wParam == SPI_SETSIPINFO) {
			if (as_SHHandleWMSettingChange)
				as_SHHandleWMSettingChange(hwnd, -1, 0, &sai); // send wParam=-1, lParam=0
		}
		else {
			if (as_SHHandleWMSettingChange)
				as_SHHandleWMSettingChange(hwnd, wParam, lParam, &sai);
		}
#endif
		break;
#ifdef USE_RELAYOUTDIALOG
	case WM_SIZE:
		RelayoutDialog(hinst, hwnd, MAKEINTRESOURCE(IsLandscape() ? IDD_L_LICENCEBOX : IDD_LICENCEBOX));
		{
			RECT rt;
			GetClientRect(hwnd, &rt);
			MoveWindow(GetDlgItem(hwnd, IDC_E_LICENCE), rt.left, rt.top, rt.right - rt.left, rt.bottom - rt.top, FALSE);
		}
		break;
#endif
	}
	return 0;
}

static INT_PTR CALLBACK AboutProc(HWND hwnd, UINT msg,
				  WPARAM wParam, LPARAM lParam)
{
	char *str;
	const UINT_PTR timerid = 0;
	const UINT elapse = 30.0 * 1000;

	debug_level((DBG_WMSG, "\nAboutProc Message %d(%x, %x)\n", msg, wParam, lParam));

	switch (msg) {
	case WM_INITDIALOG:
		debug_level((DBG_WMSG, "WM_INITDLG\n"));
#ifdef SHELL_AYGSHELL
		{
		SHINITDLGINFO shidi;
		/* Create a Done button and size it.*/
		shidi.dwMask = SHIDIM_FLAGS;
		shidi.dwFlags = SHIDIF_DONEBUTTON | SHIDIF_SIPDOWN | SHIDIF_SIZEDLGFULLSCREEN |
				SHIDIF_EMPTYMENU;
		shidi.hDlg = hwnd;
		if (as_SHInitDialog)
			as_SHInitDialog(&shidi);
		}
#endif
		//str = dupprintf("About %s", appname);
		//SetWindowTextA(hwnd, str);
		//sfree(str);
		PostMessage(hwnd, WM_COMMAND, IDC_ICON_PUTTY, 0);
		PostMessage(hwnd, WM_COMMAND, IDC_TEXT_PUTTY, 0);
		PostMessage(hwnd, WM_COMMAND, IDC_BUILD, 0);
		PostMessage(hwnd, WM_COMMAND, IDC_BATTLIFE, 0);
		SetTimer(hwnd, timerid, elapse, NULL);
		return 1;
	case WM_ACTIVATE:
		if (LOWORD(wParam) == WA_INACTIVE) {
			KillTimer(hwnd, timerid);
		}
		else {
			SetTimer(hwnd, timerid, elapse, NULL);
		}
		break;
	case WM_COMMAND:
		debug_level((DBG_WMSG, "WM_COMMAND\n"));
		switch (LOWORD(wParam)) {
		case IDOK:
			debug_level((DBG_WMSG, "About - IDOK\n"));
		case IDCANCEL:
			debug_level((DBG_WMSG, "About - IDCANCEL\n"));
			EndDialog(hwnd, TRUE);
			break;
		case IDC_SHOW_LICENCE:
			debug_level((DBG_WMSG, "About - IDC_SHOW_LICENCE\n"));

			EnableWindow(hwnd, FALSE);
			DialogBox(hinst, MAKEINTRESOURCE(IsLandscape() ? IDD_L_LICENCEBOX : IDD_LICENCEBOX),
				  hwnd, LicenceProc);
			EnableWindow(hwnd, TRUE);
			SetActiveWindow(hwnd);
			SetForegroundWindow(hwnd);
			break;
		/* Load web browser */
		case IDC_WEB:
		case IDC_POCKETPUTTY_WEB:
		case IDC_YTAARCHIVE_WEB:
			debug_level((DBG_WMSG, "IDC_WEB/IDC_POCKETPUTTY_WEB\n"));
			{
			LPSHELLEXECUTEINFO lpExecInfo = snew(SHELLEXECUTEINFO);
			ZeroMemory(lpExecInfo, sizeof(SHELLEXECUTEINFO));
			lpExecInfo->cbSize = sizeof(SHELLEXECUTEINFO);
			lpExecInfo->fMask = SEE_MASK_NOCLOSEPROCESS;
			lpExecInfo->hwnd = hwnd;
			lpExecInfo->lpVerb = TEXT("open");
			if (LOWORD(wParam) == IDC_WEB) {
				lpExecInfo->lpFile = TEXT("http://www.chiark.greenend.org.uk/~sgtatham/putty/");
			} else if (LOWORD(wParam) == IDC_YTAARCHIVE_WEB) {
				lpExecInfo->lpFile = TEXT("http://www.maroon.dti.ne.jp/yta/");
			} else {
				lpExecInfo->lpFile = TEXT("http://www.pocketputty.net/");
			}
			lpExecInfo->lpParameters = NULL;
			lpExecInfo->nShow = SW_SHOWNORMAL;
			if (ShellExecuteEx(lpExecInfo) == 0) {
				lpExecInfo->lpParameters = lpExecInfo->lpFile;
				lpExecInfo->lpFile = TEXT("iexplore.exe");
				ShellExecuteEx(lpExecInfo);
			}
			sfree(lpExecInfo);
			}
			break;
		case IDC_ICON_PUTTY:
			{
			HICON hOld = (HICON)SendDlgItemMessage(hwnd,
				LOWORD(wParam), STM_GETIMAGE, IMAGE_ICON, 0);
			HICON hNew = LoadIcon(hinst, MAKEINTRESOURCE(IDI_MAINICON));
			SendDlgItemMessage(hwnd, LOWORD(wParam),
				STM_SETIMAGE, IMAGE_ICON, (LPARAM)hNew);
			DestroyIcon(hOld);
			}
			break;
		case IDC_TEXT_PUTTY:
			SetDlgItemTextA(hwnd, IDC_TEXT_PUTTY, appname);
			break;
		case IDC_BUILD:
			SetDlgItemTextA(hwnd, IDC_BUILD, ver);
			break;
		case IDC_BATTLIFE:
			{
			SYSTEM_POWER_STATUS_EX2 powerstat;
			if (GetSystemPowerStatusEx2(&powerstat,
						    sizeof(SYSTEM_POWER_STATUS_EX2),
						    FALSE)) {
				wchar_t buf[64];
				StringCbPrintf(buf, sizeof(buf),
					       L"Battery capacity: %lu%%",
					       powerstat.BatteryLifePercent);
				SetDlgItemTextW(hwnd, IDC_BATTLIFE, buf);
			}
			}
			break;
		}
		break;
	case WM_CLOSE:
		debug_level((DBG_WMSG, "WM_CLOSE\n"));
		KillTimer(hwnd, timerid);
		EndDialog(hwnd, TRUE);
		break;
	case WM_TIMER:
		PostMessage(hwnd, WM_COMMAND, IDC_BATTLIFE, 0);
		break;
#ifdef USE_RELAYOUTDIALOG
	case WM_SIZE:
		RelayoutDialog(hinst, hwnd, MAKEINTRESOURCE(IsLandscape() ? IDD_L_ABOUTBOX : IDD_ABOUTBOX));
		PostMessage(hwnd, WM_COMMAND, IDC_ICON_PUTTY, 0);
		PostMessage(hwnd, WM_COMMAND, IDC_TEXT_PUTTY, 0);
		PostMessage(hwnd, WM_COMMAND, IDC_BUILD, 0);
		PostMessage(hwnd, WM_COMMAND, IDC_BATTLIFE, 0);
		break;
#endif
	}
	return 0;
}

int CALLBACK EnumFontFamProc(LOGFONT FAR *lpelf, TEXTMETRIC FAR *lpntm, int nFontType, LPARAM lParam)
{
	SendMessage((HWND)lParam, CB_ADDSTRING, 0, (LPARAM)lpelf->lfFaceName);
	return 1;
}

int CALLBACK SetFontCharsetProc(LOGFONT FAR *lpelf, TEXTMETRIC FAR *lpntm, int nFontType, LPARAM lParam)
{
	FontSpec *font = conf_get_fontspec(conf, CONF_font);
	font->charset = lpelf->lfCharSet;
	return 1;
}

static INT_PTR CALLBACK CfgChooseFontProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	static HWND hFontlist = NULL;
	FontSpec *font;
	HDC hdc;
	WCHAR szBuff[512];
	int i;
	debug_level((DBG_WMSG, "\nCfgChooseFontProc Message %d(%x, %x)\n", msg, wParam, lParam));

	switch (msg) {
	case WM_INITDIALOG:
		debug_level((DBG_WMSG, "WM_INITDLG\n"));
#ifdef SHELL_AYGSHELL
		{
		SHINITDLGINFO shidi;
		/* Create a Done button and size it.*/
		shidi.dwMask = SHIDIM_FLAGS;
		shidi.dwFlags = SHIDIF_DONEBUTTON | SHIDIF_SIPDOWN | SHIDIF_SIZEDLGFULLSCREEN |
				SHIDIF_EMPTYMENU;
		shidi.hDlg = hwnd;
		if (as_SHInitDialog)
			as_SHInitDialog(&shidi);
		}
#endif
		hdc = GetDC(hwnd);
		hFontlist = GetDlgItem(hwnd, IDC_FONTSPEC_NAME);
		SendMessage(hFontlist, CB_RESETCONTENT, 0, 0);
		EnumFontFamilies(hdc, NULL, (FONTENUMPROC)EnumFontFamProc, (LPARAM)hFontlist);
		ReleaseDC(hwnd, hdc);
		font = conf_get_fontspec(conf, CONF_font);
		mbstowcs(szBuff, font->name, lenof(szBuff));
		SendMessageW(hFontlist, CB_SELECTSTRING, 0, (LPARAM)szBuff);
		if (font->isbold) {
			M_CHECKDLGBUTTON(hwnd, IDC_FONTSPEC_ISBOLD, BST_CHECKED);
		} else {
			M_CHECKDLGBUTTON(hwnd, IDC_FONTSPEC_ISBOLD, BST_UNCHECKED);
		}
		_itow(font->height, szBuff, 10);
		SetDlgItemTextW(hwnd, IDC_FONTSPEC_HEIGHT, szBuff);
		SendMessage(GetDlgItem(hwnd, IDC_FONTSPEC_HEIGHTSPIN), UDM_SETRANGE32, (WPARAM)6, (LPARAM)32);
		SendMessageW(GetDlgItem(hwnd, IDC_FONTSPEC_QUALITY), CB_ADDSTRING, 0, (LPARAM)load_unicode_string(IDS_FONTSPEC_QUALITY_DEFAULT));
		SendMessageW(GetDlgItem(hwnd, IDC_FONTSPEC_QUALITY), CB_ADDSTRING, 0, (LPARAM)load_unicode_string(IDS_FONTSPEC_QUALITY_ANTIALIASED));
		SendMessageW(GetDlgItem(hwnd, IDC_FONTSPEC_QUALITY), CB_ADDSTRING, 0, (LPARAM)load_unicode_string(IDS_FONTSPEC_QUALITY_NON_ANTIALIASED));
		SendMessageW(GetDlgItem(hwnd, IDC_FONTSPEC_QUALITY), CB_ADDSTRING, 0, (LPARAM)load_unicode_string(IDS_FONTSPEC_QUALITY_CLEARTYPE));
		SendMessage(GetDlgItem(hwnd, IDC_FONTSPEC_QUALITY), CB_SETCURSEL, (WPARAM)conf_get_int(conf, CONF_font_quality), 0);
		break;
	case WM_COMMAND:
		debug_level((DBG_WMSG, "WM_COMMAND\n"));
		switch (LOWORD(wParam)) {
		case IDOK:
			debug_level((DBG_WMSG, "IDOK\n"));
			i = SendMessage(hFontlist, CB_GETCURSEL, 0, 0);
			if (i != CB_ERR) {
				char *name;
				SendMessageW(hFontlist, CB_GETLBTEXT, (WPARAM)i, (LPARAM)szBuff);
				name = unicode2ansi(szBuff);
				hdc = GetDC(hwnd);
				EnumFontFamilies(hdc, szBuff, (FONTENUMPROC)SetFontCharsetProc, 0);
				ReleaseDC(hwnd, hdc);
				GetDlgItemTextW(hwnd, IDC_FONTSPEC_HEIGHT, szBuff, lenof(szBuff));
				font = fontspec_new(name,
						    (M_ISDLGBUTTONCHECKED(hwnd, IDC_FONTSPEC_ISBOLD) == BST_CHECKED) ? 1 : 0,
						    _wtoi(szBuff),
						    DEFAULT_CHARSET);
				conf_set_int(conf, CONF_font_quality, SendMessage(GetDlgItem(hwnd, IDC_FONTSPEC_QUALITY), CB_GETCURSEL, 0, 0));
				conf_set_fontspec(conf, CONF_font, font);
				sfree(name);
			}
			EndDialog(hwnd, TRUE);
			return 1;
		case IDCANCEL:
			debug_level((DBG_WMSG, "IDCANCEL\n"));
			EndDialog(hwnd, FALSE);
			return 1;
		}
		break;
	case WM_CLOSE:
		debug_level((DBG_WMSG, "WM_CLOSE\n"));
		EndDialog(hwnd, FALSE);
		break;
#ifdef USE_RELAYOUTDIALOG
	case WM_SIZE:
		RelayoutDialog(hinst, hwnd, MAKEINTRESOURCE(IsLandscape() ? IDD_L_CFG_CHOOSEFONT : IDD_CFG_CHOOSEFONT));
		break;
#endif
	}
	return 0;
}

int choosefont(HWND hwnd)
{
	return DialogBox(hinst, MAKEINTRESOURCE(IsLandscape() ? IDD_L_CFG_CHOOSEFONT : IDD_CFG_CHOOSEFONT),
			 hwnd, CfgChooseFontProc);
}

#ifdef USE_PROPERTYSHEET
static HWND hCfgProperty = NULL;
static int PropertySheetCommand(HWND hwnd, WPARAM wParam, LPARAM lParam);
#endif

static HWND hCfgSession = NULL, hCfgWindow = NULL, hCfgSize = NULL, hCfgSsh = NULL,
	    hCfgData = NULL, hCfgConnection = NULL, hCfgSerial = NULL;
#if !(UNDER_CE < 0x400)
static HWND hCfgTunnel = NULL;
#endif
enum { CFG_TAB_SESSION, CFG_TAB_WINDOW, CFG_TAB_SIZE, CFG_TAB_CONNECTION, CFG_TAB_DATA, CFG_TAB_SSH,
#if !(UNDER_CE < 0x400)
      CFG_TAB_TUNNEL,
#endif
      CFG_TAB_SERIAL, CFG_TAB_MAX };

static INT_PTR CALLBACK CfgSessionProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	WCHAR szBuff[1024];
	static HWND hSesslist = NULL, hProtocol = NULL;
	struct sesslist sesslist;
	char *sessionname;
	int isdef, i;
	debug_level((DBG_WMSG, "\nCfgSessionProc Message %d(%x, %x)\n", msg, wParam, lParam));

	switch (msg) {
	case WM_INITDIALOG:
		debug_level((DBG_WMSG, "WM_INITDLG\n"));
#ifdef USE_PROPERTYSHEET
		if (!hCfgSession)
			hCfgSession = hwnd;
#endif
		if (conf_get_int(conf, CONF_protocol) == PROT_SERIAL) {
			SetDlgItemTextA(hwnd, IDC_HOST, conf_get_str(conf, CONF_serline));
			_itow(conf_get_int(conf, CONF_serspeed), szBuff, 10);
			SetDlgItemTextW(hwnd, IDC_PORT, szBuff);
		} else {
			SetDlgItemTextA(hwnd, IDC_HOST, conf_get_str(conf, CONF_host));
			_itow(conf_get_int(conf, CONF_port), szBuff, 10);
			SetDlgItemTextW(hwnd, IDC_PORT, szBuff);
		}
		hProtocol = GetDlgItem(hwnd, IDC_PROTOCOL);
		SendMessage(hProtocol, CB_RESETCONTENT, 0, 0);
		SendMessageW(hProtocol, CB_ADDSTRING, 0, (LPARAM)load_unicode_string(IDS_PROTOCOL_RAW));
		SendMessageW(hProtocol, CB_ADDSTRING, 0, (LPARAM)load_unicode_string(IDS_PROTOCOL_TELNET));
		SendMessageW(hProtocol, CB_ADDSTRING, 0, (LPARAM)load_unicode_string(IDS_PROTOCOL_RLOGIN));
		SendMessageW(hProtocol, CB_ADDSTRING, 0, (LPARAM)load_unicode_string(IDS_PROTOCOL_SSH));
		SendMessageW(hProtocol, CB_ADDSTRING, 0, (LPARAM)load_unicode_string(IDS_PROTOCOL_SERIAL));
		SendMessage(hProtocol, CB_SETCURSEL, (WPARAM)conf_get_int(conf, CONF_protocol), 0);
		if (conf_get_int(conf, CONF_protocol) == PROT_SERIAL) {
			SetDlgItemTextW(hwnd, IDC_TEXT_HOST, load_unicode_string(IDS_TEXT_HOST_SERIALLINE));
			SetDlgItemTextW(hwnd, IDC_TEXT_PORT, load_unicode_string(IDS_TEXT_PORT_SPEED));
		} else {
			SetDlgItemTextW(hwnd, IDC_TEXT_HOST, load_unicode_string(IDS_TEXT_HOST_HOSTNAME));
			SetDlgItemTextW(hwnd, IDC_TEXT_PORT, load_unicode_string(IDS_TEXT_PORT_PORT));
		}
		get_sesslist(&sesslist, TRUE);
		hSesslist = GetDlgItem(hwnd, IDC_SESSLIST);
		SendMessage(hSesslist, LB_RESETCONTENT, 0, 0);
		for (i = 0; i < sesslist.nsessions; i++) {
			mbstowcs(szBuff, sesslist.sessions[i], lenof(szBuff));
			SendMessageW(hSesslist, LB_ADDSTRING, 0, (LPARAM)szBuff);
		}
		get_sesslist(&sesslist, FALSE);
		break;
	case WM_COMMAND:
		debug_level((DBG_WMSG, "WM_COMMAND\n"));
		switch (LOWORD(wParam)) {
		case IDC_HOST:
			if (HIWORD(wParam) == EN_UPDATE) {
				char *str;
				GetDlgItemTextW(hwnd, IDC_HOST, szBuff, lenof(szBuff));
				str = unicode2ansi(szBuff);
				if (conf_get_int(conf, CONF_protocol) == PROT_SERIAL) {
					conf_set_str(conf, CONF_serline, str);
				} else {
					conf_set_str(conf, CONF_host, str);
				}
				sfree(str);
				return 1;
			}
			break;
		case IDC_PORT:
			if (HIWORD(wParam) == EN_UPDATE) {
				GetDlgItemTextW(hwnd, IDC_PORT, szBuff, lenof(szBuff));
				if (conf_get_int(conf, CONF_protocol) == PROT_SERIAL) {
					conf_set_int(conf, CONF_serspeed, _wtoi(szBuff));
				} else {
					conf_set_int(conf, CONF_port, _wtoi(szBuff));
				}
				return 1;
			}
			break;
		case IDC_PROTOCOL:
			if (HIWORD(wParam) == CBN_SELCHANGE) {
				conf_set_int(conf, CONF_protocol, SendMessage(hProtocol, CB_GETCURSEL, 0, 0));
				switch (conf_get_int(conf, CONF_protocol)) {
				case PROT_RAW:
					SetDlgItemTextA(hwnd, IDC_HOST, conf_get_str(conf, CONF_host));
					SetDlgItemTextA(hwnd, IDC_PORT, "");
					SetDlgItemTextW(hwnd, IDC_TEXT_HOST, load_unicode_string(IDS_TEXT_HOST_HOSTNAME));
					SetDlgItemTextW(hwnd, IDC_TEXT_PORT, load_unicode_string(IDS_TEXT_PORT_PORT));
					break;
				case PROT_TELNET:
					SetDlgItemTextA(hwnd, IDC_HOST, conf_get_str(conf, CONF_host));
					SetDlgItemTextA(hwnd, IDC_PORT, "23");
					SetDlgItemTextW(hwnd, IDC_TEXT_HOST, load_unicode_string(IDS_TEXT_HOST_HOSTNAME));
					SetDlgItemTextW(hwnd, IDC_TEXT_PORT, load_unicode_string(IDS_TEXT_PORT_PORT));
					break;
				case PROT_RLOGIN:
					SetDlgItemTextA(hwnd, IDC_HOST, conf_get_str(conf, CONF_host));
					SetDlgItemTextA(hwnd, IDC_PORT, "513");
					SetDlgItemTextW(hwnd, IDC_TEXT_HOST, load_unicode_string(IDS_TEXT_HOST_HOSTNAME));
					SetDlgItemTextW(hwnd, IDC_TEXT_PORT, load_unicode_string(IDS_TEXT_PORT_PORT));
					break;
				case PROT_SSH:
					SetDlgItemTextA(hwnd, IDC_HOST, conf_get_str(conf, CONF_host));
					SetDlgItemTextA(hwnd, IDC_PORT, "22");
					SetDlgItemTextW(hwnd, IDC_TEXT_HOST, load_unicode_string(IDS_TEXT_HOST_HOSTNAME));
					SetDlgItemTextW(hwnd, IDC_TEXT_PORT, load_unicode_string(IDS_TEXT_PORT_PORT));
					break;
				case PROT_SERIAL:
					SetDlgItemTextA(hwnd, IDC_HOST, conf_get_str(conf, CONF_serline));
					_itow(conf_get_int(conf, CONF_serspeed), szBuff, 10);
					SetDlgItemTextW(hwnd, IDC_PORT, szBuff);
					SetDlgItemTextW(hwnd, IDC_TEXT_HOST, load_unicode_string(IDS_TEXT_HOST_SERIALLINE));
					SetDlgItemTextW(hwnd, IDC_TEXT_PORT, load_unicode_string(IDS_TEXT_PORT_SPEED));
					break;
				}
			}
			break;
		case IDC_SESSION_LOAD:
			if ((i = SendMessage(hSesslist, LB_GETCURSEL, 0, 0)) < 0) break;
			SendMessageW(hSesslist, LB_GETTEXT, (WPARAM)i, (LPARAM)szBuff);
			sessionname = unicode2ansi(szBuff);
			if (!strcmp(sessionname, "")) {
				sfree(sessionname);
				break;
			}
			isdef = !strcmp(sessionname, "Default Settings");
			load_settings(sessionname, conf);
			sfree(sessionname);
			if (isdef) {
				SetDlgItemTextW(hwnd, IDC_SESSION, L"");
			} else {
				SetDlgItemTextW(hwnd, IDC_SESSION, szBuff);
			}
			SendMessage(hCfgSession, WM_INITDIALOG, 0, 0);
			SendMessage(hCfgWindow, WM_INITDIALOG, 0, 0);
			SendMessage(hCfgSize, WM_INITDIALOG, 0, 0);
			SendMessage(hCfgConnection, WM_INITDIALOG, 0, 0);
			SendMessage(hCfgData, WM_INITDIALOG, 0, 0);
			SendMessage(hCfgSsh, WM_INITDIALOG, 0, 0);
#if !(UNDER_CE < 0x400)
			SendMessage(hCfgTunnel, WM_INITDIALOG, 0, 0);
#endif
			SendMessage(hCfgSerial, WM_INITDIALOG, 0, 0);
			break;
		case IDC_SESSION_SAVE:
			GetDlgItemTextW(hwnd, IDC_SESSION, szBuff, lenof(szBuff));
			sessionname = unicode2ansi(szBuff);
			if (!strcmp(sessionname, "")) {
				sfree(sessionname);
				break;
			}
			isdef = !strcmp(sessionname, "Default Settings");
			save_settings(sessionname, conf);
			get_sesslist(&sesslist, TRUE);
			SendMessage(hSesslist, LB_RESETCONTENT, 0, 0);
			for (i = 0; i < sesslist.nsessions; i++) {
				mbstowcs(szBuff, sesslist.sessions[i], lenof(szBuff));
				SendMessageW(hSesslist, LB_ADDSTRING, 0, (LPARAM)szBuff);
			}
			get_sesslist(&sesslist, FALSE);
			sfree(sessionname);
			break;
		case IDC_SESSION_DEL:
			if ((i = SendMessage(hSesslist, LB_GETCURSEL, 0, 0)) < 0) break;
			SendMessage(hSesslist, LB_GETTEXT, (WPARAM)i, (LPARAM)szBuff);
			sessionname = unicode2ansi(szBuff);
			del_settings(sessionname);
			get_sesslist(&sesslist, TRUE);
			SendMessage(hSesslist, LB_RESETCONTENT, 0, 0);
			for (i = 0; i < sesslist.nsessions; i++) {
				mbstowcs(szBuff, sesslist.sessions[i], lenof(szBuff));
				SendMessageW(hSesslist, LB_ADDSTRING, 0, (LPARAM)szBuff);
			}
			get_sesslist(&sesslist, FALSE);
			sfree(sessionname);
			break;
		case IDC_SESSLIST:
			if (HIWORD(wParam) == LBN_DBLCLK) {
				i = SendMessage(hSesslist, LB_GETCURSEL, 0, 0);
				SendMessage(hSesslist, LB_GETTEXT, (WPARAM)i, (LPARAM)szBuff);
				sessionname = unicode2ansi(szBuff);
				isdef = !strcmp(sessionname, "Default Settings");
				load_settings(sessionname, conf);
				sfree(sessionname);
				if (strcmp(conf_get_str(conf, CONF_host), "")) {
#ifdef USE_PROPERTYSHEET
					SendMessage(GetParent(hwnd), WM_COMMAND, (WPARAM)IDOK, 0);
#else
					SendMessage(GetParent(GetParent(hwnd)), WM_COMMAND, (WPARAM)IDOK, 0);
#endif
				} else {
					if (isdef) {
						SetDlgItemTextW(hwnd, IDC_SESSION, L"");
					} else {
						SetDlgItemTextW(hwnd, IDC_SESSION, szBuff);
					}
					SendMessage(hCfgSession, WM_INITDIALOG, 0, 0);
					SendMessage(hCfgWindow, WM_INITDIALOG, 0, 0);
					SendMessage(hCfgSize, WM_INITDIALOG, 0, 0);
					SendMessage(hCfgConnection, WM_INITDIALOG, 0, 0);
					SendMessage(hCfgData, WM_INITDIALOG, 0, 0);
					SendMessage(hCfgSsh, WM_INITDIALOG, 0, 0);
#if !(UNDER_CE < 0x400)
					SendMessage(hCfgTunnel, WM_INITDIALOG, 0, 0);
#endif
					SendMessage(hCfgSerial, WM_INITDIALOG, 0, 0);
				}
			}
			break;
#ifdef USE_PROPERTYSHEET
		default:
			PropertySheetCommand(hwnd, wParam, lParam);
			break;
#endif
		}
		break;
	case WM_NOTIFY:
		debug_level((DBG_WMSG, "WM_NOTIFY\n"));
		switch (((NMHDR *)lParam)->code) {
		case NM_SETFOCUS:
			if (((NMHDR *)lParam)->hwndFrom == GetDlgItem(hwnd, IDC_HOST)) {
				HIMC himc = ImmGetContext(hwnd);
				if (himc) {
					ImmSetOpenStatus(himc, FALSE);
					ImmReleaseContext(hwnd, himc);
				}
			}
			break;
		}
		break;
#ifdef USE_RELAYOUTDIALOG
	case WM_SIZE:
		RelayoutDialog(hinst, hwnd, MAKEINTRESOURCE(IsLandscape() ? IDD_L_CFG_SESSION : IDD_CFG_SESSION));
		break;
#endif
	}
	return 0;
}

static INT_PTR CALLBACK CfgWindowProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	WCHAR szBuff[1024];
	char szBuff2[1024];
	const char *cp;
	static HWND hCodepage = NULL;
	FontSpec *font;
	int i, j;
	debug_level((DBG_WMSG, "\nCfgWindowProc Message %d(%x, %x)\n", msg, wParam, lParam));
	switch (msg) {
	case WM_INITDIALOG:
		debug_level((DBG_WMSG, "WM_INITDLG\n"));
#ifdef USE_PROPERTYSHEET
		if (!hCfgWindow)
			hCfgWindow = hwnd;
#endif
		_itow(conf_get_int(conf, CONF_savelines), szBuff, 10);
		SetDlgItemTextW(hwnd, IDC_SAVELINES, szBuff);
		if (conf_get_int(conf, CONF_scrollbar)) {
			M_CHECKDLGBUTTON(hwnd, IDC_SCROLLBAR, BST_CHECKED);
		} else {
			M_CHECKDLGBUTTON(hwnd, IDC_SCROLLBAR, BST_UNCHECKED);
		}
		if (conf_get_int(conf, CONF_scrollbar_in_fullscreen)) {
			M_CHECKDLGBUTTON(hwnd, IDC_SCROLLBAR_IN_FULLSCREEN, BST_CHECKED);
		} else {
			M_CHECKDLGBUTTON(hwnd, IDC_SCROLLBAR_IN_FULLSCREEN, BST_UNCHECKED);
		}
		PostMessage(hwnd, WM_COMMAND, IDC_FONTSPEC, 0);
		hCodepage = GetDlgItem(hwnd, IDC_CODEPAGE);
		SendMessage(hCodepage, CB_RESETCONTENT, 0, 0);
		j = -1;
		for (i = 0, cp = cp_enumerate(i); cp != NULL; i++, cp = cp_enumerate(i)) {
			mbstowcs(szBuff, cp, lenof(szBuff));
			SendMessageW(hCodepage, CB_ADDSTRING, 0, (LPARAM)szBuff);
			if (!strcmp(conf_get_str(conf, CONF_line_codepage), cp)) j = i;
		}
		if (j < 0) {
			SendMessage(hCodepage, CB_SETCURSEL, (WPARAM)(i -1), 0);
		} else {
			SendMessage(hCodepage, CB_SETCURSEL, (WPARAM)j, 0);
		}
		break;
	case WM_COMMAND:
		debug_level((DBG_WMSG, "WM_COMMAND\n"));
		switch (LOWORD(wParam)) {
		case IDC_SAVELINES:
			if (HIWORD(wParam) == EN_UPDATE) {
				GetDlgItemTextW(hwnd, IDC_SAVELINES, szBuff, lenof(szBuff));
				conf_set_int(conf, CONF_savelines, _wtoi(szBuff));
				return 1;
			}
			break;
		case IDC_SCROLLBAR:
			conf_set_int(conf, CONF_scrollbar, (M_ISDLGBUTTONCHECKED(hwnd, IDC_SCROLLBAR) == BST_CHECKED));
			break;
		case IDC_SCROLLBAR_IN_FULLSCREEN:
			conf_set_int(conf, CONF_scrollbar_in_fullscreen, (M_ISDLGBUTTONCHECKED(hwnd, IDC_SCROLLBAR_IN_FULLSCREEN) == BST_CHECKED));
			break;
		case IDC_FONTSPEC:
			font = conf_get_fontspec(conf, CONF_font);
			if (font->isbold) {
				sprintf(szBuff2, "%s, %s, %dpt", font->name, load_ansi_string(IDS_FONTSPEC_QUALITY_BOLD), font->height);
			} else {
				sprintf(szBuff2, "%s, %dpt", font->name, font->height);
			}
			switch (conf_get_int(conf, CONF_font_quality)) {
			case FQ_ANTIALIASED:
				strcat(szBuff2, ", ");
				strcat(szBuff2, load_ansi_string(IDS_FONTSPEC_QUALITY_ANTIALIASED));
				break;
			case FQ_NONANTIALIASED:
				strcat(szBuff2, ", ");
				strcat(szBuff2, load_ansi_string(IDS_FONTSPEC_QUALITY_NON_ANTIALIASED));
				break;
			case FQ_CLEARTYPE:
				strcat(szBuff2, ", ");
				strcat(szBuff2, load_ansi_string(IDS_FONTSPEC_QUALITY_CLEARTYPE));
				break;
			}
			SetDlgItemTextA(hwnd, IDC_FONTSPEC, szBuff2);
			break;
		case IDC_CHOOSEFONT:
			if (choosefont(hwnd)) {
				PostMessage(hwnd, WM_COMMAND, IDC_FONTSPEC, 0);
			}
			break;
		case IDC_CODEPAGE:
			if (HIWORD(wParam) == CBN_SELCHANGE) {
				cp = cp_enumerate(SendMessage(GetDlgItem(hwnd, IDC_CODEPAGE), CB_GETCURSEL, 0, 0));
				if (cp == NULL || !strcmp(cp, "Use font encoding")) {
					strcpy(szBuff2, "");
				} else {
					strcpy(szBuff2, cp);
				}
				conf_set_str(conf, CONF_line_codepage, szBuff2);
			}
			break;
#ifdef USE_PROPERTYSHEET
		default:
			PropertySheetCommand(hwnd, wParam, lParam);
			break;
#endif
		}
		break;
#ifdef USE_RELAYOUTDIALOG
	case WM_SIZE:
		RelayoutDialog(hinst, hwnd, MAKEINTRESOURCE(IsLandscape() ? IDD_L_CFG_WINDOW : IDD_CFG_WINDOW));
		PostMessage(hwnd, WM_COMMAND, IDC_FONTSPEC, 0);
		break;
#endif
	}
	return 0;
}

static INT_PTR CALLBACK CfgSizeProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	WCHAR szBuff[1024];
	char szBuff2[1024];
	debug_level((DBG_WMSG, "\CfgSizeProc Message %d(%x, %x)\n", msg, wParam, lParam));
	switch (msg) {
	case WM_INITDIALOG:
		debug_level((DBG_WMSG, "WM_INITDLG\n"));
#ifdef USE_PROPERTYSHEET
		if (!hCfgSize)
			hCfgSize = hwnd;
#endif
		_itow(conf_get_int(conf, CONF_height), szBuff, 10);
		SetDlgItemTextW(hwnd, IDC_SIZE_HEIGHT, szBuff);
		_itow(conf_get_int(conf, CONF_width), szBuff, 10);
		SetDlgItemTextW(hwnd, IDC_SIZE_WIDTH, szBuff);
		switch (conf_get_int(conf, CONF_resize_action)) {
		case RESIZE_TERM:
			M_CHECKDLGBUTTON(hwnd, IDC_SIZETYPE_TERM, BST_CHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_SIZETYPE_FONT, BST_UNCHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_SIZETYPE_EITHER, BST_UNCHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_SIZETYPE_DISABLED, BST_UNCHECKED);
			break;
		case RESIZE_FONT:
			M_CHECKDLGBUTTON(hwnd, IDC_SIZETYPE_TERM, BST_UNCHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_SIZETYPE_FONT, BST_CHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_SIZETYPE_EITHER, BST_UNCHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_SIZETYPE_DISABLED, BST_UNCHECKED);
			break;
		case RESIZE_EITHER:
			M_CHECKDLGBUTTON(hwnd, IDC_SIZETYPE_TERM, BST_UNCHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_SIZETYPE_FONT, BST_UNCHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_SIZETYPE_EITHER, BST_CHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_SIZETYPE_DISABLED, BST_UNCHECKED);
			break;
		case RESIZE_DISABLED:
			M_CHECKDLGBUTTON(hwnd, IDC_SIZETYPE_TERM, BST_UNCHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_SIZETYPE_FONT, BST_UNCHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_SIZETYPE_EITHER, BST_UNCHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_SIZETYPE_DISABLED, BST_CHECKED);
			break;
		}
		break;
	case WM_COMMAND:
		debug_level((DBG_WMSG, "WM_COMMAND\n"));
		switch (LOWORD(wParam)) {
		case IDC_SIZE_HEIGHT:
			if (HIWORD(wParam) == EN_UPDATE) {
				GetDlgItemTextW(hwnd, IDC_SIZE_HEIGHT, szBuff, lenof(szBuff));
				conf_set_int(conf, CONF_height, _wtoi(szBuff));
				return 1;
			}
			break;
		case IDC_SIZE_WIDTH:
			if (HIWORD(wParam) == EN_UPDATE) {
				GetDlgItemTextW(hwnd, IDC_SIZE_WIDTH, szBuff, lenof(szBuff));
				conf_set_int(conf, CONF_width, _wtoi(szBuff));
				return 1;
			}
			break;
		case IDC_SIZETYPE_TERM:
			conf_set_int(conf, CONF_resize_action, RESIZE_TERM);
			break;
		case IDC_SIZETYPE_FONT:
			conf_set_int(conf, CONF_resize_action, RESIZE_FONT);
			break;
		case IDC_SIZETYPE_EITHER:
			conf_set_int(conf, CONF_resize_action, RESIZE_EITHER);
			break;
		case IDC_SIZETYPE_DISABLED:
			conf_set_int(conf, CONF_resize_action, RESIZE_DISABLED);
			break;
#ifdef USE_PROPERTYSHEET
		default:
			PropertySheetCommand(hwnd, wParam, lParam);
			break;
#endif
		}
		break;
#ifdef USE_RELAYOUTDIALOG
	case WM_SIZE:
		RelayoutDialog(hinst, hwnd, MAKEINTRESOURCE(IsLandscape() ? IDD_L_CFG_SIZE : IDD_CFG_SIZE));
		break;
#endif
	}
	return 0;
}

static INT_PTR CALLBACK CfgConnectionProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	WCHAR szBuff[512];
	debug_level((DBG_WMSG, "\nCfgConnectionProc Message %d(%x, %x)\n", msg, wParam, lParam));
	switch (msg) {
	case WM_INITDIALOG:
		debug_level((DBG_WMSG, "WM_INITDLG\n"));
#ifdef USE_PROPERTYSHEET
		if (!hCfgConnection)
			hCfgConnection = hwnd;
#endif
		_itow(conf_get_int(conf, CONF_ping_interval), szBuff, 10);
		SetDlgItemTextW(hwnd, IDC_PING_INTERVAL, szBuff);
		if (conf_get_int(conf, CONF_tcp_nodelay)) {
			M_CHECKDLGBUTTON(hwnd, IDC_TCP_NODELAY, BST_CHECKED);
		} else {
			M_CHECKDLGBUTTON(hwnd, IDC_TCP_NODELAY, BST_UNCHECKED);
		}
		if (conf_get_int(conf, CONF_tcp_keepalives)) {
			M_CHECKDLGBUTTON(hwnd, IDC_TCP_KEEPALIVE, BST_CHECKED);
		} else {
			M_CHECKDLGBUTTON(hwnd, IDC_TCP_KEEPALIVE, BST_UNCHECKED);
		}
#ifdef NO_IPV6
		conf_set_int(conf, CONF_addressfamily, ADDRTYPE_IPV4);
		M_CHECKDLGBUTTON(hwnd, IDC_ADDRTYPE_UNSPEC, BST_UNCHECKED);
		M_CHECKDLGBUTTON(hwnd, IDC_ADDRTYPE_IPV4, BST_CHECKED);
		M_CHECKDLGBUTTON(hwnd, IDC_ADDRTYPE_IPV6, BST_UNCHECKED);
		EnableWindow(GetDlgItem(hwnd, IDC_ADDRTYPE_UNSPEC), FALSE);
		EnableWindow(GetDlgItem(hwnd, IDC_ADDRTYPE_IPV4), FALSE);
		EnableWindow(GetDlgItem(hwnd, IDC_ADDRTYPE_IPV6), FALSE);
#else
		switch (conf_get_int(conf, CONF_addressfamily)) {
		case ADDRTYPE_UNSPEC:
			M_CHECKDLGBUTTON(hwnd, IDC_ADDRTYPE_UNSPEC, BST_CHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_ADDRTYPE_IPV4, BST_UNCHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_ADDRTYPE_IPV6, BST_UNCHECKED);
			break;
		case ADDRTYPE_IPV4:
			M_CHECKDLGBUTTON(hwnd, IDC_ADDRTYPE_UNSPEC, BST_UNCHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_ADDRTYPE_IPV4, BST_CHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_ADDRTYPE_IPV6, BST_UNCHECKED);
			break;
		case ADDRTYPE_IPV6:
			M_CHECKDLGBUTTON(hwnd, IDC_ADDRTYPE_UNSPEC, BST_UNCHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_ADDRTYPE_IPV4, BST_UNCHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_ADDRTYPE_IPV6, BST_CHECKED);
			break;
		}
#endif
		break;
	case WM_COMMAND:
		debug_level((DBG_WMSG, "WM_COMMAND\n"));
		switch (LOWORD(wParam)) {
		case IDC_PING_INTERVAL:
			if (HIWORD(wParam) == EN_UPDATE) {
				GetDlgItemTextW(hwnd, IDC_PING_INTERVAL, szBuff, lenof(szBuff));
				conf_set_int(conf, CONF_ping_interval, _wtoi(szBuff));
			}
			break;
		case IDC_TCP_NODELAY:
			conf_set_int(conf, CONF_tcp_nodelay,
				     (M_ISDLGBUTTONCHECKED(hwnd, IDC_TCP_NODELAY) == BST_CHECKED));
			break;
		case IDC_TCP_KEEPALIVE:
			conf_set_int(conf, CONF_tcp_keepalives,
				     (M_ISDLGBUTTONCHECKED(hwnd, IDC_TCP_KEEPALIVE) == BST_CHECKED));
			break;
		case IDC_ADDRTYPE_UNSPEC:
			conf_set_int(conf, CONF_addressfamily, ADDRTYPE_UNSPEC);
			break;
		case IDC_ADDRTYPE_IPV4:
			conf_set_int(conf, CONF_addressfamily, ADDRTYPE_IPV4);
			break;
		case IDC_ADDRTYPE_IPV6:
			conf_set_int(conf, CONF_addressfamily, ADDRTYPE_IPV6);
			break;
#ifdef USE_PROPERTYSHEET
		default:
			PropertySheetCommand(hwnd, wParam, lParam);
			break;
#endif
		}
		break;
#ifdef USE_RELAYOUTDIALOG
	case WM_SIZE:
		RelayoutDialog(hinst, hwnd, MAKEINTRESOURCE(IsLandscape() ? IDD_L_CFG_CONNECTION : IDD_CFG_CONNECTION));
		break;
#endif
	}
	return 0;
}

static INT_PTR CALLBACK CfgDataProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	WCHAR szBuff[1024];
	char szBuff2[1024];
	debug_level((DBG_WMSG, "\nCfgDataProc Message %d(%x, %x)\n", msg, wParam, lParam));
	switch (msg) {
	case WM_INITDIALOG:
		debug_level((DBG_WMSG, "WM_INITDLG\n"));
#ifdef USE_PROPERTYSHEET
		if (!hCfgData)
			hCfgData = hwnd;
#endif
		SetDlgItemTextA(hwnd, IDC_USERNAME, conf_get_str(conf, CONF_username));
		SetDlgItemTextA(hwnd, IDC_TERMTYPE, conf_get_str(conf, CONF_termtype));
		SetDlgItemTextA(hwnd, IDC_TERMSPEED, conf_get_str(conf, CONF_termspeed));
		break;
	case WM_COMMAND:
		debug_level((DBG_WMSG, "WM_COMMAND\n"));
		switch (LOWORD(wParam)) {
		case IDC_USERNAME:
			if (HIWORD(wParam) == EN_UPDATE) {
				GetDlgItemTextA(hwnd, IDC_USERNAME, szBuff2, lenof(szBuff2));
				szBuff2[lenof(szBuff2) - 1] = '\0';

				conf_set_str(conf, CONF_username, szBuff2);
			}
			break;
		case IDC_TERMTYPE:
			if (HIWORD(wParam) == EN_UPDATE) {
				GetDlgItemTextA(hwnd, IDC_TERMTYPE, szBuff2, lenof(szBuff2));
				szBuff2[lenof(szBuff2) - 1] = '\0';
				conf_set_str(conf, CONF_termtype, szBuff2);
			}
			break;
		case IDC_TERMSPEED:
			if (HIWORD(wParam) == EN_UPDATE) {
				GetDlgItemTextA(hwnd, IDC_TERMSPEED, szBuff2, lenof(szBuff2));
				szBuff2[lenof(szBuff2) - 1] = '\0';
				conf_set_str(conf, CONF_termspeed, szBuff2);
			}
			break;
#ifdef USE_PROPERTYSHEET
		default:
			PropertySheetCommand(hwnd, wParam, lParam);
			break;
#endif
		}
		break;
	case WM_NOTIFY:
		debug_level((DBG_WMSG, "WM_NOTIFY\n"));
		switch (((NMHDR *)lParam)->code) {
		case NM_SETFOCUS:
			if (((NMHDR *)lParam)->hwndFrom == GetDlgItem(hwnd, IDC_USERNAME) ||
			    ((NMHDR *)lParam)->hwndFrom == GetDlgItem(hwnd, IDC_TERMTYPE) ||
			    ((NMHDR *)lParam)->hwndFrom == GetDlgItem(hwnd, IDC_TERMSPEED)) {
				HIMC himc = ImmGetContext(hwnd);
				if (himc) {
					ImmSetOpenStatus(himc, FALSE);
					ImmReleaseContext(hwnd, himc);
				}
			}
			break;
		}
		break;
#ifdef USE_RELAYOUTDIALOG
	case WM_SIZE:
		RelayoutDialog(hinst, hwnd, MAKEINTRESOURCE(IsLandscape() ? IDD_L_CFG_DATA : IDD_CFG_DATA));
		break;
#endif
	}
	return 0;
}

static INT_PTR CALLBACK CfgSshProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	WCHAR szBuff[1024];
	debug_level((DBG_WMSG, "\nCfgSshProc Message %d(%x, %x)\n", msg, wParam, lParam));

	switch (msg) {
	case WM_INITDIALOG:
		debug_level((DBG_WMSG, "WM_INITDLG\n"));
#ifdef USE_PROPERTYSHEET
		if (!hCfgSsh)
			hCfgSsh = hwnd;
#endif
		if (conf_get_int(conf, CONF_compression)) {
			M_CHECKDLGBUTTON(hwnd, IDC_COMPRESSION, BST_CHECKED);
		} else {
			M_CHECKDLGBUTTON(hwnd, IDC_COMPRESSION, BST_UNCHECKED);
		}
		switch (conf_get_int(conf, CONF_sshprot)) {
		case 0:
			M_CHECKDLGBUTTON(hwnd, IDC_PROT_SSH1ONLY, BST_CHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_PROT_SSH1, BST_UNCHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_PROT_SSH2, BST_UNCHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_PROT_SSH2ONLY, BST_UNCHECKED);
			break;
		case 1:
			M_CHECKDLGBUTTON(hwnd, IDC_PROT_SSH1ONLY, BST_UNCHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_PROT_SSH1, BST_CHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_PROT_SSH2, BST_UNCHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_PROT_SSH2ONLY, BST_UNCHECKED);
			break;
		case 2:
			M_CHECKDLGBUTTON(hwnd, IDC_PROT_SSH1ONLY, BST_UNCHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_PROT_SSH1, BST_UNCHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_PROT_SSH2, BST_CHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_PROT_SSH2ONLY, BST_UNCHECKED);
			break;
		case 3:
			M_CHECKDLGBUTTON(hwnd, IDC_PROT_SSH1ONLY, BST_UNCHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_PROT_SSH1, BST_UNCHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_PROT_SSH2, BST_UNCHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_PROT_SSH2ONLY, BST_CHECKED);
			break;
		}
		SetDlgItemTextA(hwnd, IDC_KEYFILE, filename_to_str(conf_get_filename(conf, CONF_keyfile)));
		break;
	case WM_COMMAND:
		debug_level((DBG_WMSG, "WM_COMMAND\n"));
		switch (LOWORD(wParam)) {
		case IDC_COMPRESSION:
			conf_set_int(conf, CONF_compression, (M_ISDLGBUTTONCHECKED(hwnd, IDC_COMPRESSION) == BST_CHECKED));
			break;
		case IDC_PROT_SSH1ONLY:
			conf_set_int(conf, CONF_sshprot, 0);
			break;
		case IDC_PROT_SSH1:
			conf_set_int(conf, CONF_sshprot, 1);
			break;
		case IDC_PROT_SSH2:
			conf_set_int(conf, CONF_sshprot, 2);
			break;
		case IDC_PROT_SSH2ONLY:
			conf_set_int(conf, CONF_sshprot, 3);
			break;
		case IDC_CHOOSEKEYFILE:
			{
			OPENFILENAME ofn;
			WCHAR szFile[FILENAME_MAX];
			HINSTANCE hInstOFN = LoadLibrary(TEXT("gsGetFile.dll"));
			const char afilters[] = FILTER_KEY_FILES;
			WCHAR *filters;
			unsigned l;
			l = MultiByteToWideChar(CP_ACP, 0, afilters, lenof(afilters), NULL, 0);
			filters = snewn(l + 2, WCHAR);
			MultiByteToWideChar(CP_ACP, 0, afilters, lenof(afilters), filters, l + 2);
			filters[l + 1] = filters[l] = L'\0';
			ZeroMemory(&ofn, sizeof(ofn));
			ofn.lStructSize = sizeof(ofn);
			ofn.hwndOwner = hwnd;
			ofn.lpstrFile = szFile;
			ofn.nMaxFile = FILENAME_MAX;
			ofn.lpstrFilter = filters;
			ofn.lpstrTitle = load_unicode_string(IDS_PRIVATE_KEY_DIALOG);
			ofn.Flags = OFN_FILEMUSTEXIST;
			szFile[0] = L'\0';
			if (hInstOFN) {
				BOOL (*gsGetOpenFileName)(OPENFILENAME *pofn);
#ifdef __GNUC__
				gsGetOpenFileName = GetProcAddress(hInstOFN, TEXT("gsGetOpenFileName"));
#else
				(FARPROC)gsGetOpenFileName = GetProcAddress(hInstOFN, TEXT("gsGetOpenFileName"));
#endif
				if (gsGetOpenFileName) {
					mbstowcs(szFile, filename_to_str(conf_get_filename(conf, CONF_keyfile)), lenof(szFile));
					if (gsGetOpenFileName(&ofn)) SetDlgItemTextW(hwnd, IDC_KEYFILE, szFile);
				} else {
					if (GetOpenFileName(&ofn)) SetDlgItemTextW(hwnd, IDC_KEYFILE, szFile);
				}
				FreeLibrary(hInstOFN);
			} else {
				if (GetOpenFileName(&ofn)) SetDlgItemTextW(hwnd, IDC_KEYFILE, szFile);
			}
			sfree(filters);
			}
			break;
		case IDC_KEYFILE:
			if (HIWORD(wParam) == EN_UPDATE) {
				char szFile[FILENAME_MAX];
				Filename *keyfile;
				GetDlgItemTextA(hwnd, IDC_KEYFILE, szFile, lenof(szFile));
				keyfile = filename_from_str(szFile);
				conf_set_filename(conf, CONF_keyfile, keyfile);
				filename_free(keyfile);
				return 1;
			}
			break;
#ifdef USE_PROPERTYSHEET
		default:
			PropertySheetCommand(hwnd, wParam, lParam);
			break;
#endif
		}
		break;
#ifdef USE_RELAYOUTDIALOG
	case WM_SIZE:
		RelayoutDialog(hinst, hwnd, MAKEINTRESOURCE(IsLandscape() ? IDD_L_CFG_SSH : IDD_CFG_SSH));
		break;
#endif
	}
	return 0;
}

#if !(UNDER_CE < 0x400)
#define WM_USER_RECONF (WM_USER + 1)
static INT_PTR CALLBACK CfgTunnelProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	static HWND hPfwdList = NULL;
	WCHAR szBuff[512], src[16], dest[512];
	static WCHAR type;
	char *p, szBuff2[512];
	char *key, *val;
	int i, j;
	debug_level((DBG_WMSG, "\nCfgTunnelProc Message %d(%x, %x)\n", msg, wParam, lParam));
	switch (msg) {
	case WM_INITDIALOG:
		debug_level((DBG_WMSG, "WM_INITDLG\n"));
#ifdef USE_PROPERTYSHEET
		if (!hCfgTunnel)
			hCfgTunnel = hwnd;
#endif
		SetDlgItemTextA(hwnd, IDC_PFWD_SRC, "");
		SetDlgItemTextA(hwnd, IDC_PFWD_DEST, "");
		M_CHECKDLGBUTTON(hwnd, IDC_PFWD_LOCAL, BST_CHECKED);
		M_CHECKDLGBUTTON(hwnd, IDC_PFWD_REMOTE, BST_UNCHECKED);
		M_CHECKDLGBUTTON(hwnd, IDC_PFWD_DYNAMIC, BST_UNCHECKED);
		type = 'L';
		hPfwdList = GetDlgItem(hwnd, IDC_PFWD_LIST);
		PostMessage(hwnd, WM_USER_RECONF, 0, 0);
		break;
	case WM_USER_RECONF:
		SendMessage(hPfwdList, LB_RESETCONTENT, 0, 0);
		for (val = conf_get_str_strs(conf, CONF_portfwd, NULL, &key);
		     val != NULL;
		     val = conf_get_str_strs(conf, CONF_portfwd, key, &key)) {
			char *p;
        		if (!strcmp(val, "D")) {
				char *L;
				/*

				 * A dynamic forwarding is stored as L12345=D or
				 * 6L12345=D (since it's mutually exclusive with
				 * L12345=anything else), but displayed as D12345
				 * to match the fiction that 'Local', 'Remote' and
				 * 'Dynamic' are three distinct modes and also to

				 * align with OpenSSH's command line option syntax
				 * that people will already be used to. So, for
				 * display purposes, find the L in the key string
				 * and turn it into a D.

				 */
				p = dupprintf("%s\t", key);
				L = strchr(p, 'L');
				if (L) *L = 'D';
			} else
				p = dupprintf("%s\t%s", key, val);
			mbstowcs(szBuff, p, lenof(szBuff));
			SendMessageW(hPfwdList, LB_ADDSTRING, 0, (LPARAM)szBuff);
			sfree(p);
		}
		break;
	case WM_COMMAND:
		debug_level((DBG_WMSG, "WM_COMMAND\n"));
		switch (LOWORD(wParam)) {
		case IDC_PFWD_LOCAL:
			type = 'L';
			break;
		case IDC_PFWD_REMOTE:
			type = 'R';
			break;
		case IDC_PFWD_DYNAMIC:
			type = 'D';
			break;
		case IDC_PFWD_ADD:
			src[0] = type;
			GetDlgItemTextW(hwnd, IDC_PFWD_SRC, src +1, lenof(src) -1);
			GetDlgItemTextW(hwnd, IDC_PFWD_DEST, dest, lenof(dest));
			if (wcslen(src) == 1) {
				char *cap = dupprintf(load_ansi_string(IDS_WARN_CAPTION_ERROR), appname);
				MessageBoxA(hwnd, load_ansi_string(IDS_WARN_NEED_SOURCE_PORT_NUMBER),
					    cap, MB_OK|MB_ICONERROR);
				sfree(cap);
				break;
			}
			if (type != 'D' && wcsstr(dest, TEXT(":")) == NULL) {
				char *cap = dupprintf(load_ansi_string(IDS_WARN_CAPTION_ERROR), appname);
				MessageBoxA(hwnd, load_ansi_string(IDS_WARN_NEED_DESTINATION_ADDRESS),
					    cap, MB_OK|MB_ICONERROR);
				sfree(cap);
				break;
			}
			if (type == 'D') {
				src[0] = 'L';
			}
			wcscpy(szBuff, src);
			key = unicode2ansi(szBuff);
			if (type == 'D') {
				wcscpy(szBuff, L"D");
			} else {
				wcscpy(szBuff, dest);
			}
			val = unicode2ansi(szBuff);
			if (conf_get_str_str_opt(conf, CONF_portfwd, key)) {
				char *cap = dupprintf(load_ansi_string(IDS_WARN_CAPTION_ERROR), appname);
				MessageBoxA(hwnd, load_ansi_string(IDS_WARN_FORWARDING_ALREADY_EXISTS),
					    cap, MB_OK|MB_ICONERROR);
				sfree(cap);
			} else {
				conf_set_str_str(conf, CONF_portfwd, key, val);
			}

			sfree(key);
			sfree(val);

			PostMessage(hwnd, WM_USER_RECONF, 0, 0);
			SetDlgItemTextA(hwnd, IDC_PFWD_SRC, "");
			SetDlgItemTextA(hwnd, IDC_PFWD_DEST, "");
			break;
		case IDC_PFWD_REMOVE:
			if ((i = SendMessage(hPfwdList, LB_GETCURSEL, 0, 0)) < 0) break;
			SendMessage(hPfwdList, LB_DELETESTRING, (WPARAM)i, 0);
			j = SendMessage(hPfwdList, LB_GETCOUNT, 0, 0);

			key = conf_get_str_nthstrkey(conf, CONF_portfwd, i);
			val = conf_get_str_str(conf, CONF_portfwd, key);
			type = key[0];
			if (!strcmp(val, "D")) {
				type = 'D';
				val = "";
			}

			p = key + 1;
			mbstowcs(src, p, lenof(src));
			mbstowcs(dest, val, lenof(dest));

			M_CHECKDLGBUTTON(hwnd, IDC_PFWD_LOCAL, type == 'L' ? BST_CHECKED : BST_UNCHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_PFWD_REMOTE, type == 'R' ? BST_CHECKED : BST_UNCHECKED);
			M_CHECKDLGBUTTON(hwnd, IDC_PFWD_DYNAMIC, type == 'D' ? BST_CHECKED : BST_UNCHECKED);
			SetDlgItemTextW(hwnd, IDC_PFWD_SRC, src);
			SetDlgItemTextW(hwnd, IDC_PFWD_DEST, dest);
			/* And delete it */
			conf_del_str_str(conf, CONF_portfwd, key);

			PostMessage(hwnd, WM_USER_RECONF, 0, 0);
			break;
#ifdef USE_PROPERTYSHEET
		default:
			PropertySheetCommand(hwnd, wParam, lParam);
			break;
#endif
		}
		break;
	case WM_NOTIFY:
		debug_level((DBG_WMSG, "WM_NOTIFY\n"));
		switch (((NMHDR *)lParam)->code) {
		case NM_SETFOCUS:
			if (((NMHDR *)lParam)->hwndFrom == GetDlgItem(hwnd, IDC_PFWD_DEST)) {
				HIMC himc = ImmGetContext(hwnd);
				if (himc) {
					ImmSetOpenStatus(himc, FALSE);
					ImmReleaseContext(hwnd, himc);
				}
			}
			break;
		}
		break;
#ifdef USE_RELAYOUTDIALOG
	case WM_SIZE:
		RelayoutDialog(hinst, hwnd, MAKEINTRESOURCE(IsLandscape() ? IDD_L_CFG_TUNNEL : IDD_CFG_TUNNEL));
		break;
#endif
	}
	return 0;
}
#undef WM_USER_RECONF
#endif

static INT_PTR CALLBACK CfgSerialProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	WCHAR szBuff[512];
	static HWND hParity = NULL, hFlow = NULL;
	debug_level((DBG_WMSG, "\nCfgSerialProc Message %d(%x, %x)\n", msg, wParam, lParam));
	switch (msg) {
	case WM_INITDIALOG:
		debug_level((DBG_WMSG, "WM_INITDLG\n"));
#ifdef USE_PROPERTYSHEET
		if (!hCfgSerial)
			hCfgSerial = hwnd;
#endif
		_itow(conf_get_int(conf, CONF_serdatabits), szBuff, 10);
		SetDlgItemTextW(hwnd, IDC_SER_DATABITS, szBuff);
		_itow(conf_get_int(conf, CONF_serstopbits), szBuff, 10);
		SetDlgItemTextW(hwnd, IDC_SER_STOPBITS, szBuff);
		hParity = GetDlgItem(hwnd, IDC_SER_PARITY);
		SendMessage(hParity, CB_RESETCONTENT, 0, 0);
		SendMessageW(hParity, CB_ADDSTRING, 0, (LPARAM)load_unicode_string(IDS_SER_PARITY_NONE));
		SendMessageW(hParity, CB_ADDSTRING, 0, (LPARAM)load_unicode_string(IDS_SER_PARITY_ODD));
		SendMessageW(hParity, CB_ADDSTRING, 0, (LPARAM)load_unicode_string(IDS_SER_PARITY_EVEN));
		SendMessageW(hParity, CB_ADDSTRING, 0, (LPARAM)load_unicode_string(IDS_SER_PARITY_MASK));
		SendMessageW(hParity, CB_ADDSTRING, 0, (LPARAM)load_unicode_string(IDS_SER_PARITY_SPACE));
		SendMessage(hParity, CB_SETCURSEL, (WPARAM)conf_get_int(conf, CONF_serparity), 0);
		hFlow = GetDlgItem(hwnd, IDC_SER_FLOW);
		SendMessage(hFlow, CB_RESETCONTENT, 0, 0);
		SendMessageW(hFlow, CB_ADDSTRING, 0, (LPARAM)load_unicode_string(IDS_SER_FLOW_NONE));
		SendMessageW(hFlow, CB_ADDSTRING, 0, (LPARAM)load_unicode_string(IDS_SER_FLOW_XONXOFF));
		SendMessageW(hFlow, CB_ADDSTRING, 0, (LPARAM)load_unicode_string(IDS_SER_FLOW_RTSCTS));
		SendMessageW(hFlow, CB_ADDSTRING, 0, (LPARAM)load_unicode_string(IDS_SER_FLOW_DSRDTR));
		SendMessage(hFlow, CB_SETCURSEL, (WPARAM)conf_get_int(conf, CONF_serflow), 0);
		break;
	case WM_COMMAND:
		debug_level((DBG_WMSG, "WM_COMMAND\n"));
		switch (LOWORD(wParam)) {
		case IDC_SER_DATABITS:
			if (HIWORD(wParam) == EN_UPDATE) {
				GetDlgItemTextW(hwnd, IDC_SER_DATABITS, szBuff, lenof(szBuff));
				conf_set_int(conf, CONF_serdatabits, _wtoi(szBuff));
				return 1;
			}
			break;
		case IDC_SER_STOPBITS:
			if (HIWORD(wParam) == EN_UPDATE) {
				GetDlgItemTextW(hwnd, IDC_SER_STOPBITS, szBuff, lenof(szBuff));
				conf_set_int(conf, CONF_serstopbits, _wtoi(szBuff));
				return 1;
			}
			break;
		case IDC_SER_PARITY:
			if (HIWORD(wParam) == CBN_SELCHANGE) {
				conf_set_int(conf, CONF_serparity, SendMessage(hParity, CB_GETCURSEL, 0, 0));
				return 1;
			}
			break;
		case IDC_SER_FLOW:
			if (HIWORD(wParam) == CBN_SELCHANGE) {
				conf_set_int(conf, CONF_serflow, SendMessage(hFlow, CB_GETCURSEL, 0, 0));
				return 1;
			}
			break;
#ifdef USE_PROPERTYSHEET
		default:
			PropertySheetCommand(hwnd, wParam, lParam);
			break;
#endif
		}
		break;
#ifdef USE_RELAYOUTDIALOG
	case WM_SIZE:
		RelayoutDialog(hinst, hwnd, MAKEINTRESOURCE(IsLandscape() ? IDD_L_CFG_SERIAL : IDD_CFG_SERIAL));
		break;
#endif
	}
	return 0;
}

#ifdef USE_PROPERTYSHEET
static int PropertySheetCommand(HWND hwnd, WPARAM wParam, LPARAM lParam)
{
	debug_level((DBG_WMSG, "PropertySheetCommand Message\n"));

	switch (LOWORD(wParam)) {
		case IDOK:
		case IDM_CONF_OPEN:
			debug_level((DBG_WMSG, "IDOK\n"));
			PropSheet_PressButton(hCfgProperty, PSBTN_OK);
			return 0;
		case IDCANCEL:
		case IDM_CLOSE:
			debug_level((DBG_WMSG, "IDCANCEL\n"));
			PropSheet_PressButton(hCfgProperty, PSBTN_CANCEL);
			return 0;
		case IDM_ABOUT:
			debug_level((DBG_WMSG, "IDM_ABOUT\n"));
			showabout(hwnd);
			return 0;
		case IDM_CFG_TAB_SESSION:
			debug_level((DBG_WMSG, "IDM_CFG_TAB_SESSION\n"));
			PropSheet_SetCurSel(hCfgProperty, NULL, CFG_TAB_SESSION);
			return 0;
		case IDM_CFG_TAB_WINDOW:
			debug_level((DBG_WMSG, "IDM_CFG_TAB_WINDOW\n"));
			PropSheet_SetCurSel(hCfgProperty, NULL, CFG_TAB_WINDOW);
			return 0;
		case IDM_CFG_TAB_SIZE:
			debug_level((DBG_WMSG, "IDM_CFG_TAB_SIZE\n"));
			PropSheet_SetCurSel(hCfgProperty, NULL, CFG_TAB_SIZE);
			return 0;
		case IDM_CFG_TAB_CONNECTION:
			debug_level((DBG_WMSG, "IDM_CFG_TAB_CONNECTION\n"));
			PropSheet_SetCurSel(hCfgProperty, NULL, CFG_TAB_CONNECTION);
			return 0;
		case IDM_CFG_TAB_DATA:
			debug_level((DBG_WMSG, "IDM_CFG_TAB_DATA\n"));
			PropSheet_SetCurSel(hCfgProperty, NULL, CFG_TAB_DATA);
			return 0;
		case IDM_CFG_TAB_SSH:
			debug_level((DBG_WMSG, "IDM_CFG_TAB_SSH\n"));
			PropSheet_SetCurSel(hCfgProperty, NULL, CFG_TAB_SSH);
			return 0;
#if !(UNDER_CE < 0x400)
		case IDM_CFG_TAB_TUNNEL:
			debug_level((DBG_WMSG, "IDM_CFG_TAB_TUNNEL\n"));
			PropSheet_SetCurSel(hCfgProperty, NULL, CFG_TAB_TUNNEL);
			return 0;
#endif
		case IDM_CFG_TAB_SERIAL:
			debug_level((DBG_WMSG, "IDM_CFG_TAB_SERIAL\n"));
			PropSheet_SetCurSel(hCfgProperty, NULL, CFG_TAB_SERIAL);
			return 0;
	}
	return 1;
}

static INT_PTR CALLBACK PropSheetInitProc(HWND hwnd, UINT msg, LPARAM lParam)
{
	debug_level((DBG_WMSG, "PropSheetInitProc Message %d(%x)\n", msg, lParam));

	switch (msg) {
	case PSCB_PRECREATE:
		if (0xFFFF == ((LPDLGTEMPLATEEX)lParam)->signature) {
			((LPDLGTEMPLATEEX)lParam)->style &= ~(DS_CONTEXTHELP);
		}
		else {
			((LPDLGTEMPLATE)lParam)->style &= ~(DS_CONTEXTHELP);
		}
		return 1;
	case PSCB_INITIALIZED:
		hCfgProperty = hwnd;
#ifdef SHELL_AYGSHELL
		{
		SHINITDLGINFO shidi;
		/* Create a Cancel button and size it.*/
		shidi.dwMask = SHIDIM_FLAGS;
		shidi.dwFlags = SHIDIF_CANCELBUTTON | SHIDIF_SIPDOWN | SHIDIF_SIZEDLGFULLSCREEN |
				SHIDIF_WANTSCROLLBAR;
		shidi.hDlg = hwnd;
		if (as_SHInitDialog)
			as_SHInitDialog(&shidi);
		if (as_SHFullScreen)
			as_SHFullScreen(hwnd, SHFS_SHOWSIPBUTTON);
		if (as_SHDoneButton)
			as_SHDoneButton(hwnd, SHDB_SHOWCANCEL);
		}
#endif
		SetWindowTextW(hwnd, load_unicode_string(IDS_CFG_CAPTION));
		CreateRpCommandBar(hwnd);
		{
		HWND hCfgTab = PropSheet_GetTabControl(hwnd);
		RECT rt;
		GetWindowRect(hCfgTab, &rt);
		rt.top += override_menu_height;
		SetWindowPos(hCfgTab, NULL,
			     rt.left, rt.top, rt.right - rt.left, rt.bottom - rt.top,
			     SWP_NOZORDER);
		}
#if !(UNDER_CE < 0x400)
		EnableMenuItem(popup_menu, IDM_CFG_TAB_TUNNEL, MF_ENABLED | MF_BYCOMMAND);
#endif
		SetWindowLongPtr(hwnd, GWL_STYLE, GetWindowLongPtr(hwnd, GWL_STYLE) | WS_MINIMIZEBOX);
		break;
	case PSCB_GETVERSION:
		/* Windows Common Controls Version */
		return COMCTL32_VERSION;
	}
	return 0;
}

static INT_PTR CALLBACK PropSheetPageProc(HWND hwnd, UINT msg, LPPROPSHEETPAGE ppsp)
{
	debug_level((DBG_WMSG, "PropSheetPageProc Message %d\n", msg));

	switch (msg) {
	case PSPCB_RELEASE:
		break;
	case PSPCB_CREATE:
		break;
	}
	return 1;
}

static int RegisterPropertySheetPages(HINSTANCE hinst, HPROPSHEETPAGE *hPsp)
{
	PROPSHEETPAGE psp;
	WCHAR label[24];
	int i;

	debug_level((DBG_WMSG, "RegisterPropertySheetPages\n"));

	ZeroMemory(&psp, sizeof(PROPSHEETPAGE));

	psp.dwSize = sizeof(PROPSHEETPAGE);
	psp.dwFlags = PSP_DEFAULT | PSP_PREMATURE | PSP_USECALLBACK | PSP_USETITLE;
	psp.hInstance = hinst;
	psp.pfnCallback = (LPFNPSPCALLBACK)PropSheetPageProc;
	i = 0;

	psp.pszTitle = label;
	wcsncpy(label, load_unicode_string(IDS_CFG_TAB_SESSION), lenof(label));
	psp.pszTemplate = MAKEINTRESOURCE(IsLandscape() ? IDD_L_CFG_SESSION : IDD_CFG_SESSION);
	psp.pfnDlgProc = (DLGPROC)CfgSessionProc;
	hPsp[i] = CreatePropertySheetPage(&psp);
	++i;

	psp.pszTitle = label;
	wcsncpy(label, load_unicode_string(IDS_CFG_TAB_WINDOW), lenof(label));
	psp.pszTemplate = MAKEINTRESOURCE(IsLandscape() ? IDD_L_CFG_WINDOW : IDD_CFG_WINDOW);
	psp.pfnDlgProc = (DLGPROC)CfgWindowProc;
	hPsp[i] = CreatePropertySheetPage(&psp);
	++i;

	psp.pszTitle = label;
	wcsncpy(label, load_unicode_string(IDS_CFG_TAB_SIZE), lenof(label));
	psp.pszTemplate = MAKEINTRESOURCE(IsLandscape() ? IDD_L_CFG_SIZE : IDD_CFG_SIZE);
	psp.pfnDlgProc = (DLGPROC)CfgSizeProc;
	hPsp[i] = CreatePropertySheetPage(&psp);
	++i;

	psp.pszTitle = label;
	wcsncpy(label, load_unicode_string(IDS_CFG_TAB_CONNECTION), lenof(label));
	psp.pszTemplate = MAKEINTRESOURCE(IsLandscape() ? IDD_L_CFG_CONNECTION : IDD_CFG_CONNECTION);
	psp.pfnDlgProc = (DLGPROC)CfgConnectionProc;
	hPsp[i] = CreatePropertySheetPage(&psp);
	++i;

	psp.pszTitle = label;
	wcsncpy(label, load_unicode_string(IDS_CFG_TAB_DATA), lenof(label));
	psp.pszTemplate = MAKEINTRESOURCE(IsLandscape() ? IDD_L_CFG_DATA : IDD_CFG_DATA);
	psp.pfnDlgProc = (DLGPROC)CfgDataProc;
	hPsp[i] = CreatePropertySheetPage(&psp);
	++i;

	psp.pszTitle = label;
	wcsncpy(label, load_unicode_string(IDS_CFG_TAB_SSH), lenof(label));
	psp.pszTemplate = MAKEINTRESOURCE(IsLandscape() ? IDD_L_CFG_SSH : IDD_CFG_SSH);
	psp.pfnDlgProc = (DLGPROC)CfgSshProc;
	hPsp[i] = CreatePropertySheetPage(&psp);
	++i;

#if !(UNDER_CE < 0x400)
	psp.pszTitle = label;
	wcsncpy(label, load_unicode_string(IDS_CFG_TAB_TUNNEL), lenof(label));
	psp.pszTemplate = MAKEINTRESOURCE(IsLandscape() ? IDD_L_CFG_TUNNEL : IDD_CFG_TUNNEL);
	psp.pfnDlgProc = (DLGPROC)CfgTunnelProc;
	hPsp[i] = CreatePropertySheetPage(&psp);
	++i;
#endif

	psp.pszTitle = label;
	wcsncpy(label, load_unicode_string(IDS_CFG_TAB_SERIAL), lenof(label));
	psp.pszTemplate = MAKEINTRESOURCE(IsLandscape() ? IDD_L_CFG_SERIAL : IDD_CFG_SERIAL);
	psp.pfnDlgProc = (DLGPROC)CfgSerialProc;
	hPsp[i] = CreatePropertySheetPage(&psp);
	++i;

	return i;
}

static int CreateConfigurationDialog(HWND hwnd, HINSTANCE hinst)
{
	PROPSHEETHEADER psh;
	HPROPSHEETPAGE hPsp[CFG_TAB_MAX];

	debug_level((DBG_WMSG, "CreateConfigurationDialog\n"));

	ZeroMemory(&hPsp[0], sizeof(HPROPSHEETPAGE) * CFG_TAB_MAX);
	ZeroMemory(&psh, sizeof(PROPSHEETHEADER));
	psh.dwSize = sizeof(PROPSHEETHEADER);
	psh.dwFlags = PSH_DEFAULT | PSH_USECALLBACK | PSH_MAXIMIZE;
	psh.pfnCallback = (PFNPROPSHEETCALLBACK)PropSheetInitProc;
	psh.pszCaption = L"Putty Configuration";
	psh.hInstance = hinst;
	psh.hwndParent = hwnd;
	psh.nStartPage = 0;
	psh.phpage = hPsp;
	psh.nPages = RegisterPropertySheetPages(hinst, &hPsp[0]);

	if (PropertySheet(&psh) > 0) {
		if (conf_get_int(conf, CONF_protocol) == PROT_SERIAL) {
			if (strcmp(conf_get_str(conf, CONF_serline), "")) {
				return 1;
			}
		} else {
			if (strcmp(conf_get_str(conf, CONF_host), "")) {
				return 1;
			}
		}
	}
	else {
		debug_level((DBG_WMSG, "PropertySheet [cancel]: GLE=%lu\n", (ULONG)GetLastError()));
	}
	return 0;
}
#else
static INT_PTR CALLBACK ConfigurationProc(HWND hwnd, UINT msg,
					  WPARAM wParam, LPARAM lParam)
{
	TCITEM tc;
	RECT rt;
	static HWND hCfgTab = NULL;
#ifdef SHELL_AYGSHELL
	static SHACTIVATEINFO sai;
#endif

	debug_level((DBG_WMSG, "\nConfigurationProc Message %d(%x, %x)\n", msg, wParam, lParam));

	switch (msg) {
	case WM_INITDIALOG:
		debug_level((DBG_WMSG, "WM_INITDLG\n"));
		SetWindowTextW(hwnd, load_unicode_string(IDS_CFG_CAPTION));
#ifdef SHELL_AYGSHELL
		{
		SHINITDLGINFO shidi;
		/* Register */
		ZeroMemory(&sai, sizeof(sai));
		sai.cbSize = sizeof(sai);
		/* Create a Cancel button and size it.*/
		shidi.dwMask = SHIDIM_FLAGS;
		shidi.dwFlags = SHIDIF_CANCELBUTTON | SHIDIF_SIPDOWN | SHIDIF_SIZEDLGFULLSCREEN |
				SHIDIF_EMPTYMENU | SHIDIF_WANTSCROLLBAR;
		shidi.hDlg = hwnd;
		if (as_SHInitDialog)
			as_SHInitDialog(&shidi);
		if (as_SHFullScreen)
			as_SHFullScreen(hwnd, SHFS_SHOWSIPBUTTON);
		if (as_SHDoneButton)
			as_SHDoneButton(hwnd, SHDB_SHOWCANCEL);
		}
#endif
		CreateRpCommandBar(hwnd);
#if !(UNDER_CE < 0x400)
		EnableMenuItem(popup_menu, IDM_CFG_TAB_TUNNEL, MF_ENABLED | MF_BYCOMMAND);
#endif
		hCfgTab = GetDlgItem(hwnd, IDC_CFGTAB);
		SendMessage(hCfgTab, CCM_SETVERSION,
			    COMCTL32_VERSION, 0);
		GetWindowRect(hCfgTab, &rt);
		rt.top += override_menu_height;
		SetWindowPos(hCfgTab, NULL,
			     rt.left, rt.top, rt.right - rt.left, rt.bottom - rt.top,
			     SWP_NOZORDER);
		tc.mask = TCIF_TEXT;
		tc.pszText = load_unicode_string(IDS_CFG_TAB_SESSION);
		TabCtrl_InsertItem(hCfgTab, CFG_TAB_SESSION, &tc);
		tc.pszText = load_unicode_string(IDS_CFG_TAB_WINDOW);
		TabCtrl_InsertItem(hCfgTab, CFG_TAB_WINDOW, &tc);;
		tc.pszText = load_unicode_string(IDS_CFG_TAB_SIZE);
		TabCtrl_InsertItem(hCfgTab, CFG_TAB_SIZE, &tc);
		tc.pszText = load_unicode_string(IDS_CFG_TAB_CONNECTION);
		TabCtrl_InsertItem(hCfgTab, CFG_TAB_CONNECTION, &tc);
		tc.pszText = load_unicode_string(IDS_CFG_TAB_DATA);
		TabCtrl_InsertItem(hCfgTab, CFG_TAB_DATA, &tc);
		tc.pszText = load_unicode_string(IDS_CFG_TAB_SSH);
		TabCtrl_InsertItem(hCfgTab, CFG_TAB_SSH, &tc);
#if !(UNDER_CE < 0x400)
		tc.pszText = load_unicode_string(IDS_CFG_TAB_TUNNEL);
		TabCtrl_InsertItem(hCfgTab, CFG_TAB_TUNNEL, &tc);
#endif
		tc.pszText = load_unicode_string(IDS_CFG_TAB_SERIAL);
		TabCtrl_InsertItem(hCfgTab, CFG_TAB_SERIAL, &tc);

		hCfgSession = CreateDialog(hinst, MAKEINTRESOURCE(IsLandscape() ? IDD_L_CFG_SESSION : IDD_CFG_SESSION),
					   hCfgTab, (DLGPROC)CfgSessionProc);
		hCfgWindow = CreateDialog(hinst, MAKEINTRESOURCE(IsLandscape() ? IDD_L_CFG_WINDOW : IDD_CFG_WINDOW),
					  hCfgTab, (DLGPROC)CfgWindowProc);
		hCfgSize = CreateDialog(hinst, MAKEINTRESOURCE(IsLandscape() ? IDD_L_CFG_SIZE : IDD_CFG_SIZE),
					hCfgTab, (DLGPROC)CfgSizeProc);
		hCfgConnection = CreateDialog(hinst, MAKEINTRESOURCE(IsLandscape() ? IDD_L_CFG_CONNECTION : IDD_CFG_CONNECTION),
					      hCfgTab, (DLGPROC)CfgConnectionProc);
		hCfgData = CreateDialog(hinst, MAKEINTRESOURCE(IsLandscape() ? IDD_L_CFG_DATA : IDD_CFG_DATA),
					hCfgTab, (DLGPROC)CfgDataProc);
		hCfgSsh = CreateDialog(hinst, MAKEINTRESOURCE(IsLandscape() ? IDD_L_CFG_SSH : IDD_CFG_SSH),
				       hCfgTab, (DLGPROC)CfgSshProc);
#if !(UNDER_CE < 0x400)
		hCfgTunnel = CreateDialog(hinst, MAKEINTRESOURCE(IsLandscape() ? IDD_L_CFG_TUNNEL : IDD_CFG_TUNNEL),
					  hCfgTab, (DLGPROC)CfgTunnelProc);
#endif
		hCfgSerial = CreateDialog(hinst, MAKEINTRESOURCE(IsLandscape() ? IDD_L_CFG_SERIAL : IDD_CFG_SERIAL),
					  hCfgTab, (DLGPROC)CfgSerialProc);

		GetClientRect(hwnd, &rt);
		TabCtrl_AdjustRect(hCfgTab, FALSE, &rt);
		MapWindowPoints(hCfgTab, hwnd, (LPPOINT)&rt, 2);
		MoveWindow(hCfgSession, rt.left, rt.top, rt.right - rt.left, rt.bottom - rt.top, FALSE);
		MoveWindow(hCfgWindow, rt.left, rt.top, rt.right - rt.left, rt.bottom - rt.top, FALSE);
		MoveWindow(hCfgConnection, rt.left, rt.top, rt.right - rt.left, rt.bottom - rt.top, FALSE);
		MoveWindow(hCfgData, rt.left, rt.top, rt.right - rt.left, rt.bottom - rt.top, FALSE);
		MoveWindow(hCfgSsh, rt.left, rt.top, rt.right - rt.left, rt.bottom - rt.top, FALSE);
#if !(UNDER_CE < 0x400)
		MoveWindow(hCfgTunnel, rt.left, rt.top, rt.right - rt.left, rt.bottom - rt.top, FALSE);
#endif
		MoveWindow(hCfgSerial, rt.left, rt.top, rt.right - rt.left, rt.bottom - rt.top, FALSE);
		ShowWindow(hCfgSession, SW_SHOW);
		ShowWindow(hCfgWindow, SW_HIDE);
		ShowWindow(hCfgConnection, SW_HIDE);
		ShowWindow(hCfgData, SW_HIDE);
		ShowWindow(hCfgSsh, SW_HIDE);
#if !(UNDER_CE < 0x400)
		ShowWindow(hCfgTunnel, SW_HIDE);
#endif
		ShowWindow(hCfgSerial, SW_HIDE);
		break;
	case WM_COMMAND:
		debug_level((DBG_WMSG, "WM_COMMAND\n"));
		switch (LOWORD(wParam)) {
			case IDOK:
			case IDM_CONF_OPEN:
				debug_level((DBG_WMSG, "IDOK\n"));
				if (conf_get_int(conf, CONF_protocol) == PROT_SERIAL) {
					if (strcmp(conf_get_str(conf, CONF_serline), "")) EndDialog(hwnd, TRUE);
				} else {
					if (strcmp(conf_get_str(conf, CONF_host), "")) EndDialog(hwnd, TRUE);
				}
				break;
			case IDCANCEL:
			case IDM_CLOSE:
				debug_level((DBG_WMSG, "IDCANCEL\n"));
				EndDialog(hwnd, FALSE);
				break;
			case IDM_ABOUT:
				debug_level((DBG_WMSG, "IDM_ABOUT\n"));
				showabout(hwnd);
				break;
			case IDM_CFG_TAB_SESSION:
				debug_level((DBG_WMSG, "IDM_CFG_TAB_SESSION\n"));
				TabCtrl_SetCurSel(hCfgTab, CFG_TAB_SESSION);
				TabCtrl_SetCurFocus(hCfgTab, CFG_TAB_SESSION);
				break;
			case IDM_CFG_TAB_WINDOW:
				debug_level((DBG_WMSG, "IDM_CFG_TAB_WINDOW\n"));
				TabCtrl_SetCurSel(hCfgTab, CFG_TAB_WINDOW);
				TabCtrl_SetCurFocus(hCfgTab, CFG_TAB_WINDOW);
				break;
			case IDM_CFG_TAB_SIZE:
				debug_level((DBG_WMSG, "IDM_CFG_TAB_SIZE\n"));
				TabCtrl_SetCurSel(hCfgTab, CFG_TAB_SIZE);
				TabCtrl_SetCurFocus(hCfgTab, CFG_TAB_SIZE);
				break;
			case IDM_CFG_TAB_CONNECTION:
				debug_level((DBG_WMSG, "IDM_CFG_TAB_CONNECTION\n"));
				TabCtrl_SetCurSel(hCfgTab, CFG_TAB_CONNECTION);
				TabCtrl_SetCurFocus(hCfgTab, CFG_TAB_CONNECTION);
				break;
			case IDM_CFG_TAB_DATA:
				debug_level((DBG_WMSG, "IDM_CFG_TAB_DATA\n"));
				TabCtrl_SetCurSel(hCfgTab, CFG_TAB_DATA);
				TabCtrl_SetCurFocus(hCfgTab, CFG_TAB_DATA);
				break;
			case IDM_CFG_TAB_SSH:
				debug_level((DBG_WMSG, "IDM_CFG_TAB_SSH\n"));
				TabCtrl_SetCurSel(hCfgTab, CFG_TAB_SSH);
				TabCtrl_SetCurFocus(hCfgTab, CFG_TAB_SSH);
				break;
#if !(UNDER_CE < 0x400)
			case IDM_CFG_TAB_TUNNEL:
				debug_level((DBG_WMSG, "IDM_CFG_TAB_TUNNEL\n"));
				TabCtrl_SetCurSel(hCfgTab, CFG_TAB_TUNNEL);
				TabCtrl_SetCurFocus(hCfgTab, CFG_TAB_TUNNEL);
				break;
#endif
			case IDM_CFG_TAB_SERIAL:
				debug_level((DBG_WMSG, "IDM_CFG_TAB_SERIAL\n"));
				TabCtrl_SetCurSel(hCfgTab, CFG_TAB_SERIAL);
				TabCtrl_SetCurFocus(hCfgTab, CFG_TAB_SERIAL);
				break;
		}
		break;
	case WM_NOTIFY:
		debug_level((DBG_WMSG, "WM_NOTIFY\n"));
		switch (((NMHDR *)lParam)->code) {
		case TCN_SELCHANGE:
			if (((NMHDR *)lParam)->hwndFrom == hCfgTab) {
				switch (TabCtrl_GetCurSel(hCfgTab)) {
				case CFG_TAB_SESSION:
					ShowWindow(hCfgSession, SW_SHOW);
					ShowWindow(hCfgWindow, SW_HIDE);
					ShowWindow(hCfgSize, SW_HIDE);
					ShowWindow(hCfgConnection, SW_HIDE);
					ShowWindow(hCfgData, SW_HIDE);
					ShowWindow(hCfgSsh, SW_HIDE);
#if !(UNDER_CE < 0x400)
					ShowWindow(hCfgTunnel, SW_HIDE);
#endif
					ShowWindow(hCfgSerial, SW_HIDE);
					break;
				case CFG_TAB_WINDOW:
					ShowWindow(hCfgSession, SW_HIDE);
					ShowWindow(hCfgWindow, SW_SHOW);
					ShowWindow(hCfgSize, SW_HIDE);
					ShowWindow(hCfgConnection, SW_HIDE);
					ShowWindow(hCfgData, SW_HIDE);
					ShowWindow(hCfgSsh, SW_HIDE);
#if !(UNDER_CE < 0x400)
					ShowWindow(hCfgTunnel, SW_HIDE);
#endif
					ShowWindow(hCfgSerial, SW_HIDE);
					break;
				case CFG_TAB_SIZE:
					ShowWindow(hCfgSession, SW_HIDE);
					ShowWindow(hCfgWindow, SW_HIDE);
					ShowWindow(hCfgSize, SW_SHOW);
					ShowWindow(hCfgConnection, SW_HIDE);
					ShowWindow(hCfgData, SW_HIDE);
					ShowWindow(hCfgSsh, SW_HIDE);
#if !(UNDER_CE < 0x400)
					ShowWindow(hCfgTunnel, SW_HIDE);
#endif
					ShowWindow(hCfgSerial, SW_HIDE);
					break;
				case CFG_TAB_CONNECTION:
					ShowWindow(hCfgSession, SW_HIDE);
					ShowWindow(hCfgWindow, SW_HIDE);
					ShowWindow(hCfgSize, SW_HIDE);
					ShowWindow(hCfgConnection, SW_SHOW);
					ShowWindow(hCfgData, SW_HIDE);
					ShowWindow(hCfgSsh, SW_HIDE);
#if !(UNDER_CE < 0x400)
					ShowWindow(hCfgTunnel, SW_HIDE);
#endif
					ShowWindow(hCfgSerial, SW_HIDE);
					break;
				case CFG_TAB_DATA:
					ShowWindow(hCfgSession, SW_HIDE);
					ShowWindow(hCfgWindow, SW_HIDE);
					ShowWindow(hCfgSize, SW_HIDE);
					ShowWindow(hCfgConnection, SW_HIDE);
					ShowWindow(hCfgData, SW_SHOW);
					ShowWindow(hCfgSsh, SW_HIDE);
#if !(UNDER_CE < 0x400)
					ShowWindow(hCfgTunnel, SW_HIDE);
#endif
					ShowWindow(hCfgSerial, SW_HIDE);
					break;
				case CFG_TAB_SSH:
					ShowWindow(hCfgSession, SW_HIDE);
					ShowWindow(hCfgWindow, SW_HIDE);
					ShowWindow(hCfgSize, SW_HIDE);
					ShowWindow(hCfgConnection, SW_HIDE);
					ShowWindow(hCfgData, SW_HIDE);
					ShowWindow(hCfgSsh, SW_SHOW);
#if !(UNDER_CE < 0x400)
					ShowWindow(hCfgTunnel, SW_HIDE);
#endif
					ShowWindow(hCfgSerial, SW_HIDE);
					break;
#if !(UNDER_CE < 0x400)
				case CFG_TAB_TUNNEL:
					ShowWindow(hCfgSession, SW_HIDE);
					ShowWindow(hCfgWindow, SW_HIDE);
					ShowWindow(hCfgSize, SW_HIDE);
					ShowWindow(hCfgConnection, SW_HIDE);
					ShowWindow(hCfgData, SW_HIDE);
					ShowWindow(hCfgSsh, SW_HIDE);
					ShowWindow(hCfgTunnel, SW_SHOW);
					ShowWindow(hCfgSerial, SW_HIDE);
					break;
#endif
				case CFG_TAB_SERIAL:
					ShowWindow(hCfgSession, SW_HIDE);
					ShowWindow(hCfgWindow, SW_HIDE);
					ShowWindow(hCfgSize, SW_HIDE);
					ShowWindow(hCfgConnection, SW_HIDE);
					ShowWindow(hCfgData, SW_HIDE);
					ShowWindow(hCfgSsh, SW_HIDE);
#if !(UNDER_CE < 0x400)
					ShowWindow(hCfgTunnel, SW_HIDE);
#endif
					ShowWindow(hCfgSerial, SW_SHOW);
					break;
				}
			}
			break;
		}
		break;
	case WM_CLOSE:
		debug_level((DBG_WMSG, "WM_CLOSE\n"));
		EndDialog(hwnd, TRUE);
		break;
	case WM_SIZE:
		debug_level((DBG_WMSG, "WM_SIZE\n"));
		if (wParam == SIZE_MINIMIZED) {
			EndDialog(hwnd, FALSE);
			break;
		}
#ifdef USE_RELAYOUTDIALOG
		RelayoutDialog(hinst, hwnd, MAKEINTRESOURCE(IsLandscape() ? IDD_L_MAINBOX : IDD_MAINBOX));
#endif
		break;
	case WM_ACTIVATE:
		debug_level((DBG_WMSG, "WM_ACTIVATE\n"));
#ifdef SHELL_AYGSHELL
		if (as_SHHandleWMActivate)
			as_SHHandleWMActivate(hwnd, wParam, lParam, &sai, FALSE);
#endif
		break;
	case WM_SETTINGCHANGE:
		debug_level((DBG_WMSG, "WM_SETTINGCHANGE\n"));
#ifdef SHELL_AYGSHELL
		if (wParam == SPI_SETSIPINFO) {
			if (as_SHHandleWMSettingChange)
				as_SHHandleWMSettingChange(hwnd, -1, 0, &sai); // send wParam=-1, lParam=0
		}
		else {
			if (as_SHHandleWMSettingChange)
				as_SHHandleWMSettingChange(hwnd, wParam, lParam, &sai);
		}
#endif
		break;
	}
	return 0;
}
#endif

void modal_about_box(HWND hwnd)
{
}

void show_help(HWND hwnd)
{
}

void defuse_showwindow(void)
{
}

int do_config(void)
{
#ifdef USE_PROPERTYSHEET
	return CreateConfigurationDialog(NULL, hinst);
#else
	return DialogBox(hinst, MAKEINTRESOURCE(IsLandscape() ? IDD_L_MAINBOX : IDD_MAINBOX),
			 NULL, ConfigurationProc);
#endif
}

void showabout(HWND hwnd)
{
	DialogBox(hinst, MAKEINTRESOURCE(IsLandscape() ? IDD_L_ABOUTBOX : IDD_ABOUTBOX),
		  hwnd, AboutProc);
}

int verify_ssh_host_key(void *frontend, char *host, int port, char *keytype,
			char *keystr, char *fingerprint,
			void (*callback)(void *ctx, int result), void *ctx)
{
	int ret;

	/* shorter version due to small display */
#if 0
	static const char absentmsg[] =
		"The server's host key is not cached in the registry. "
		"The server's %s key fingerprint is:"
		" %s\n"
		"Yes - trust this host, No - risk, Cancel - don't trust host\n";
	static const char wrongmsg[] =
		"WARNING - POTENTIAL SECURITY BREACH!"
		"The server's host key does not match the one %s has "
		"cached in the registry. "
		"The new %s key fingerprint is: %s"
		"Yes - update %s's cache, continue"
		"No - connect, no cache update."
		"Cancel - don't trust host\n";
	static const char mbtitle[] = "%s Security Alert";
#endif

	debug_level((DBG_FUNCTION, "verify_ssh_host_key(%s);\n", fingerprint));

	/*
	* Verify the key against the registry.
	*/
	ret = verify_host_key(host, port, keytype, keystr);

	if (ret == 0)		       /* success - key matched OK */
		return 1;
	if (ret == 2) {		       /* key was different */
		int mbret;
		char *text = dupprintf(load_ansi_string(IDS_WARN_VERIFY_WRONG_HOSTKEY),
				       appname, keytype, fingerprint, appname);
		char *caption = dupprintf(load_ansi_string(IDS_WARN_CAPTION_SECURITY_ALERT), appname);
		mbret = MessageBoxA(NULL, text, caption,
				    MB_ICONWARNING | MB_YESNOCANCEL | MB_DEFBUTTON3 | MB_TOPMOST);
		assert(mbret==IDYES || mbret==IDNO || mbret==IDCANCEL);
		sfree(text);
		sfree(caption);
		if (mbret == IDYES) {
			store_host_key(host, port, keytype, keystr);
			return 1;
		} else if (mbret == IDNO) {
			return 1;
		}
	}
	if (ret == 1) {		       /* key was absent */
		int mbret;
		char *text = dupprintf(load_ansi_string(IDS_WARN_VERIFY_ABSENT_HOSTKEY),
				       keytype, fingerprint);
		char *caption = dupprintf(load_ansi_string(IDS_WARN_CAPTION_SECURITY_ALERT), appname);
		mbret = MessageBoxA(NULL, text, caption,
				    MB_ICONWARNING | MB_YESNOCANCEL | MB_DEFBUTTON3 | MB_TOPMOST);
		assert(mbret==IDYES || mbret==IDNO || mbret==IDCANCEL);
		sfree(text);
		sfree(caption);
		if (mbret == IDYES) {
			store_host_key(host, port, keytype, keystr);
			return 1;
		} else if (mbret == IDNO) {
			return 1;
		}
	}
	return 0; /* don't continue with connection */
}

/*
 * Ask whether the selected algorithm is acceptable (since it was
 * below the configured 'warn' threshold).
 */
int askalg(void *frontend, const char *algtype, const char *algname,
	   void (*callback)(void *ctx, int result), void *ctx)
{
	char *message, *title;
	int mbret;
#if 0
	static const char mbtitle[] = "%s Security Alert";
	static const char msg[] =
		"The first %s supported by the server\n"
		"is %.64s, which is below the configured\n"
		"warning threshold.\n"
		"Do you want to continue with this connection?\n";
#endif

	message = dupprintf(load_ansi_string(IDS_WARN_ASKALG_MESSAGE), algtype, algname);
	title = dupprintf(load_ansi_string(IDS_WARN_CAPTION_SECURITY_ALERT), appname);

	mbret = MessageBoxA(NULL, message, title,
			    MB_ICONWARNING | MB_YESNO | MB_DEFBUTTON2 | MB_TOPMOST);

	sfree(message);
	sfree(title);

	if (mbret == IDYES)
		return 1;
	else
		return 0;
}

/*
 * Ask whether to wipe a session log file before writing to it.
 * Returns 2 for wipe, 1 for append, 0 for cancel (don't log).
 */
int askappend(void *frontend, Filename *filename,
	      void (*callback)(void *ctx, int result), void *ctx)
{
	debug_level((DBG_FUNCTION, "askappend();\n"));
	return 0; /* cancel - don't log */
}

/*
 * Warn about the obsolescent key file format.
 *
 * Uniquely among these functions, this one does _not_ expect a
 * frontend handle. This means that if PuTTY is ported to a
 * platform which requires frontend handles, this function will be
 * an anomaly. Fortunately, the problem it addresses will not have
 * been present on that platform, so it can plausibly be
 * implemented as an empty function.
 */
void old_keyfile_warning(void)
{
#if 0
	static const char mbtitle[] = "%s Key File Warning";
	static const char message[] =
		"You are loading an SSH-2 private key which has an\n"
		"old version of the file format. This means your key\n"
		"file is not fully tamperproof. Future versions of\n"
		"%s may stop supporting this private key format,\n"
		"so we recommend you convert your key to the new\n"
		"format.\n"
		"\n"
		"You can perform this conversion by loading the key\n"
		"into PuTTYgen and then saving it again.";
#endif

	char *msg, *title;
	msg = dupprintf(load_ansi_string(IDS_WARN_OLD_KEYFILE_WARNING), appname);
	title = dupprintf(load_ansi_string(IDS_WARN_CAPTION_KEYFILE_WARNING), appname);

	MessageBoxA(NULL, msg, title, MB_OK | MB_TOPMOST);

	socket_reselect_all();

	sfree(msg);
	sfree(title);
}

static HWND CreateRpCommandBar(HWND hwnd)
{
	HWND hMenu;
#ifdef SHELL_AYGSHELL
	SHMENUBARINFO mbi;

	ZeroMemory(&mbi, sizeof(SHMENUBARINFO));
	mbi.cbSize     = sizeof(SHMENUBARINFO);
	mbi.hwndParent = hwnd;
	mbi.nToolBarId = IDM_CONF;
	mbi.hInstRes   = hinst;
	mbi.nBmpId     = 0;
	mbi.cBmpImages = 0;

	if (as_SHCreateMenuBar && as_SHCreateMenuBar(&mbi) && mbi.hwndMB) {
		popup_menu = (HMENU)SendMessage(mbi.hwndMB, SHCMBM_GETSUBMENU, 0, IDM_CONF_DROP);
		override_menu_height = 0;
		return mbi.hwndMB;
	}
#endif
	hMenu = CommandBar_Create(hinst, hwnd, 1);
	CommandBar_InsertMenubar(hMenu,  hinst, IDM_CONF, 0);
	GetWindow(hMenu, GW_CHILD);
	CommandBar_DrawMenuBar(hMenu, 0);
	popup_menu = GetSubMenu(CommandBar_GetMenu(hMenu, 0), 1);
	override_menu_height = CommandBar_Height(hMenu);

	return hMenu;
}

static BOOL IsLandscape(void)
{
	return (GetSystemMetrics(SM_CXSCREEN) > GetSystemMetrics(SM_CYSCREEN));
}

