/*
 * wince_misc.c: miscellaneous Windows-specific things
 */

#include <stdio.h>
#include <stdlib.h>
#include "putty.h"
#include "wince.h"

OSVERSIONINFO osVersion;

BOOL IsZoomed(HWND hwnd)
{
	debug_level((DBG_FUNCTION, "IsZoomed();\n"));
	return TRUE;
}

INT MessageBoxA(HWND hWnd, LPCSTR lpText, LPCSTR lpCaption, UINT uType)
{
	INT nResult;
	LPWSTR lpwcMsg = NULL, lpwcCap = NULL;

	if (!hWnd)
		uType |= MB_SETFOREGROUND;
	if (lpText)
		lpwcMsg = (LPWSTR) ansi2unicode(lpText);
	if (lpCaption)
		lpwcCap = (LPWSTR) ansi2unicode(lpCaption);

	nResult = MessageBoxW(hWnd, lpwcMsg, lpwcCap, uType);

	if (lpwcCap)
		sfree(lpwcCap);
	if (lpwcMsg)
		sfree(lpwcMsg);

	return nResult;
}

UINT WINAPI GetDlgItemTextA(HWND hWnd, INT nId, LPSTR lpszcText, INT nLength)
{
	UINT nResult;
	LPWSTR lpszText;

	lpszText = (LPWSTR) snewn(nLength + 1, WCHAR);

	nResult = GetDlgItemTextW(hWnd, nId, lpszText, nLength);

	wcstombs(lpszcText, lpszText, nLength);
	sfree(lpszText);

	return nResult;
}

BOOL WINAPI SetDlgItemTextA(HWND hWnd, INT nId, LPCSTR lpszcText)
{
	UINT nResult;
	LPWSTR lpszText;

	lpszText = ansi2unicode(lpszcText);

	nResult = SetDlgItemTextW(hWnd, nId, lpszText);

	sfree(lpszText);

	return nResult;
}

BOOL WINAPI SetWindowTextA(HWND hWnd, LPCSTR lpszcText)
{
	BOOL fResult;
	LPWSTR lpszText;

	lpszText = ansi2unicode(lpszcText);

	fResult = SetWindowTextW(hWnd, lpszText);

	sfree(lpszText);

	return fResult;
}

DWORD GetCurrentDirectoryW(DWORD dwSize, LPWSTR lpBuffer)
{
	WCHAR module[MAX_PATH];

	if (GetModuleFileNameW(NULL, module, MAX_PATH)) {
		module[wcsrchr(module, L'\\') - module + 1] = L'\0';
		wcsncpy(lpBuffer, module, dwSize);
	}

	return wcslen(lpBuffer);
}

UINT GetWindowsDirectoryW(LPWSTR lpBuffer, UINT uSize)
{
	wcsncpy(lpBuffer, WINCE_WIN_PATH, uSize);
	return (wcslen(WINCE_WIN_PATH));
}

BOOL GetUserNameA(LPSTR lpBuffer, LPDWORD nSize)
{
	return (FALSE);
}

BOOL GetSystemTimeAdjustment(PDWORD lpTimeAdjustment,
			     PDWORD lpTimeIncrement,
			     PBOOL lpTimeAdjustmentDisabled)
{
	*lpTimeAdjustment = 0;
	*lpTimeIncrement = 0;
	*lpTimeAdjustmentDisabled = TRUE;

	return TRUE;
}

BOOL GetProcessTimes(HANDLE hProcess, LPFILETIME lpCreationTime,
		     LPFILETIME lpExitTime, LPFILETIME lpKernelTime,
		     LPFILETIME lpUserTime)
{
	return TRUE;
}

LONG RegCreateKey(HKEY hKey, LPCTSTR lpSubKey, PHKEY phkResult)
{
	DWORD dwDisp;
	return (RegCreateKeyEx(hKey, lpSubKey, 0, NULL, 0, 0, NULL, phkResult, &dwDisp));
}

LONG RegOpenKey(HKEY hKey, LPCTSTR lpSubKey, PHKEY phkResult)
{
	return (RegOpenKeyEx(hKey, lpSubKey, 0, 0, phkResult));
}

LONG RegEnumKey(HKEY hKey, DWORD dwIndex, LPTSTR lpName, DWORD cchName)
{
	return (RegEnumKeyEx(hKey, dwIndex, lpName, &cchName, 0, NULL, NULL, NULL));
}

BOOL TextOutW(HDC hdc, int nXStart, int nYStart, LPCWSTR lpwString, int wcbString)
{
	RECT *rect = snew(RECT);
	int result;

	rect->left = nXStart;
	rect->top = nYStart;
	/* Aleq: ooops! Nasty temporary workaroung */
	rect->right = 0;
	rect->bottom = 0;

	/* Aleq: Possibly DT_NOCLIP usage? */
	rect->bottom = rect->top + DrawText(hdc, lpwString, wcbString, rect, DT_CALCRECT);
	result = DrawText(hdc, lpwString, wcbString, rect, DT_LEFT);
	sfree(rect);

	return (result);
}

HFONT WINAPI CreateFontA(int nHeight, int nWidth, int nEscapement, int nOrientation, int fnWeight,
			 DWORD fdwItalic, DWORD fdwUnderline, DWORD fdwStrikeOut, DWORD fdwCharSet,
			 DWORD fdwOutputPrecision, DWORD fdwClipPrecision, DWORD fdwQuality, DWORD fdwPitchAndFamily,
			 LPCSTR lpszFace)
{
	LOGFONTW lfw;
	wchar_t *wface;

	lfw.lfHeight = nHeight;
	lfw.lfWidth = nWidth;
	lfw.lfEscapement = nEscapement;
	lfw.lfOrientation = nOrientation;
	lfw.lfWeight = fnWeight;
	lfw.lfItalic = (BYTE) fdwItalic;
	lfw.lfUnderline = (BYTE) fdwUnderline;
	lfw.lfStrikeOut = (BYTE) fdwStrikeOut;
	lfw.lfCharSet = (BYTE) fdwCharSet;
	lfw.lfOutPrecision = (BYTE) fdwOutputPrecision;
	lfw.lfClipPrecision = (BYTE) fdwClipPrecision;
	lfw.lfQuality = (BYTE) fdwQuality;
	lfw.lfPitchAndFamily = (BYTE) fdwPitchAndFamily;

	wface = ansi2unicode(lpszFace);
	wcsncpy(lfw.lfFaceName, wface, wcslen(wface));
	sfree(wface);

	return (CreateFontIndirect(&lfw));
}

BOOL WINAPI GetTextExtentExPointA(HDC hdc, LPCSTR lpszStr, int cchString,
				  int nMaxExtent, LPINT lpnFit, LPINT alpDx, LPSIZE lpSize)
{
	BOOL fResult;
	LPWSTR lpwcStr;

	lpwcStr = ansi2unicode(lpszStr);
	fResult = GetTextExtentExPointW(hdc, lpwcStr, cchString,
					nMaxExtent, lpnFit, alpDx, lpSize);
	sfree(lpwcStr);

	return fResult;
}

BOOL GetKeyboardState(PBYTE lpKeyState)
{
	int vk;

	for (vk = 0; vk <= 0xFF; vk++)
		lpKeyState[vk] = (BYTE) GetKeyState(vk);

	return TRUE;
}

int ToAsciiEx(UINT uVirtKey, UINT uScanCode, PBYTE lpKeyState, LPWORD lpChar, UINT uFlags, HKL dwhkl)
{
	return 0;
}

BOOL init_winver(void)
{
	ZeroMemory(&osVersion, sizeof(osVersion));
	osVersion.dwOSVersionInfoSize = sizeof (OSVERSIONINFO);

	return GetVersionEx ( (OSVERSIONINFO *) &osVersion);
}

HMODULE load_system32_dll(const char *libname)
{
	/*
	 * Wrapper function to load a DLL out of c:\windows\system32
	 * without going through the full DLL search path. (Hence no
	 * attack is possible by placing a substitute DLL earlier on that
	 * path.)
	 */
	static WCHAR *sysdir = NULL;
	WCHAR *fullpath, *wlibname;
	HMODULE ret;

	if (!sysdir) {
		int size = 0, len;
		do {
			size = 3*size/2 + 512;
			sysdir = sresize(sysdir, size, WCHAR);
			len = GetWindowsDirectoryW(sysdir, size);
		} while (len >= size);
	}

	wlibname = ansi2unicode(libname);
	fullpath = snewn(wcslen(sysdir) + 1 + wcslen(wlibname) + 1, WCHAR);
	wcscpy(fullpath, sysdir);
	wcscat(fullpath, L"\\");
	wcscat(fullpath, wlibname);
	ret = LoadLibraryW(fullpath);
	sfree(fullpath);
	sfree(wlibname);
	return ret;
}

/*
 * A tree234 containing mappings from system error codes to strings.
 */

struct errstring {
	int error;
	char *text;
};

static int errstring_find(void *av, void *bv)
{
	int *a = (int *)av;
	struct errstring *b = (struct errstring *)bv;
	if (*a < b->error)
		return -1;
	if (*a > b->error)
		return +1;
	return 0;
}
static int errstring_compare(void *av, void *bv)
{
	struct errstring *a = (struct errstring *)av;
	return errstring_find(&a->error, bv);
}

static tree234 *errstrings = NULL;

const char *win_strerror(int error)
{
	struct errstring *es;

	if (!errstrings)
		errstrings = newtree234(errstring_compare);

	es = find234(errstrings, &error, errstring_find);

	if (!es) {
		WCHAR msgtext[65536]; /* maximum size for FormatMessage is 64K */
		char *ansimsg;

		es = snew(struct errstring);
		es->error = error;
		if (!FormatMessage((FORMAT_MESSAGE_FROM_SYSTEM |
				    FORMAT_MESSAGE_IGNORE_INSERTS), NULL, error,
				   MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
				   msgtext, lenof(msgtext)-1, NULL)) {
			swprintf(msgtext,
				 L"(unable to format: FormatMessage returned %d)",
				 GetLastError());
		} else {
			int len = wcslen(msgtext);
			if (len > 0 && msgtext[len-1] == L'\n')
				msgtext[len-1] = L'\0';
		}
		ansimsg = unicode2ansi(msgtext);
		es->text = dupprintf("Error %d: %s", error, ansimsg);
		sfree(ansimsg);
		add234(errstrings, es);
	}

	return es->text;
}

BOOL get_wince_oeminfo(wchar_t *oeminfo, int bufsize)
{
	return SystemParametersInfo(SPI_GETOEMINFO, bufsize, oeminfo, FALSE);
}

BOOL get_wince_platform(wchar_t *platform, int bufsize)
{
	return SystemParametersInfo(SPI_GETPLATFORMTYPE, bufsize, platform, FALSE);
}

char *unicode2ansi(const wchar_t *source)
{
	static const long safe_length = 32;
	long l;
	char *result;
#if defined(__GNUC__) && !defined(__COREDLL__)
	char *locale, *ctype;
	ctype = setlocale(LC_CTYPE, "");
	locale = setlocale(LC_ALL, ctype);
#endif
	l = wcstombs(0, source, 0);
	result = snewn(l + 1 + safe_length, char);
	wcstombs(result, source, l + 1 + safe_length);
#if defined(__GNUC__) && !defined(__COREDLL__)
	setlocale(LC_ALL, locale);
	setlocale(LC_CTYPE, ctype);
#endif
	return (result);
}

wchar_t *ansi2unicode(const char *source)
{
	static const long safe_length = 32;
	long l;
	wchar_t *result;
#if defined(__GNUC__) && !defined(__COREDLL__)
	char *locale, *ctype;
	ctype = setlocale(LC_CTYPE, "");
	locale = setlocale(LC_ALL, ctype);
#endif
	l = mbstowcs(0, source, 0);
	result = snewn(l + 1 + safe_length, wchar_t);
	mbstowcs(result, source, l + 1 + safe_length);
#if defined(__GNUC__) && !defined(__COREDLL__)
	setlocale(LC_ALL, locale);
	setlocale(LC_CTYPE, ctype);
#endif
	return (result);
}

Filename *filename_from_str(const char *str)
{
	Filename *ret = snew(Filename);
	ret->path = dupstr(str);
	return ret;
}

Filename *filename_copy(const Filename *fn)
{
	return filename_from_str(fn->path);
}

const char *filename_to_str(const Filename *fn)
{
	return fn->path;
}

int filename_equal(const Filename *f1, const Filename *f2)
{
	return !strcmp(f1->path, f2->path);
}

int filename_is_null(const Filename *fn)
{
	return !*fn->path;
}

void filename_free(Filename *fn)
{
	sfree(fn->path);
	sfree(fn);
}

int filename_serialise(const Filename *f, void *vdata)
{
	char *data = (char *)vdata;
	int len = strlen(f->path) + 1;     /* include trailing NUL */
	if (data) {
		strcpy(data, f->path);
	}
	return len;
}

Filename *filename_deserialise(void *vdata, int maxsize, int *used)
{
	char *data = (char *)vdata;
	char *end;
	end = memchr(data, '\0', maxsize);
	if (!end)
		return NULL;
	end++;
	*used = end - data;
	return filename_from_str(data);
}

char *get_username(void)
{
	DWORD namelen;
	char *user;

	namelen = 0;
	if (GetUserNameA(NULL, &namelen) == FALSE) {
		/*
		 * Apparently this doesn't work at least on Windows XP SP2.
		 * Thus assume a maximum of 256. It will fail again if it
		 * doesn't fit.
		 */
		namelen = 256;
	}

	user = snewn(namelen, char);
	user[0] = '\0';
	GetUserNameA(user, &namelen);

	return user;
}

FontSpec *fontspec_new(const char *name,
		       int bold, int height, int charset)
{
	FontSpec *f = snew(FontSpec);
	f->name = dupstr(name);
	f->isbold = bold;
	f->height = height;
	f->charset = charset;
	return f;
}

FontSpec *fontspec_copy(const FontSpec *f)
{
	return fontspec_new(f->name, f->isbold, f->height, f->charset);
}

void fontspec_free(FontSpec *f)
{
	sfree(f->name);
	sfree(f);
}

int fontspec_serialise(FontSpec *f, void *vdata)
{
	char *data = (char *)vdata;
	int len = strlen(f->name) + 1;     /* include trailing NUL */
	if (data) {
		strcpy(data, f->name);
		PUT_32BIT_MSB_FIRST(data + len, f->isbold);
		PUT_32BIT_MSB_FIRST(data + len + 4, f->height);
		PUT_32BIT_MSB_FIRST(data + len + 8, f->charset);
	}
	return len + 12;                   /* also include three 4-byte ints */
}

FontSpec *fontspec_deserialise(void *vdata, int maxsize, int *used)
{
	char *data = (char *)vdata;
	char *end;
	if (maxsize < 13)
		return NULL;
	end = memchr(data, '\0', maxsize-12);
	if (!end)
		return NULL;
	end++;
	*used = end - data + 12;
	return fontspec_new(data,
			    GET_32BIT_MSB_FIRST(end),
			    GET_32BIT_MSB_FIRST(end + 4),
			    GET_32BIT_MSB_FIRST(end + 8));
}

const wchar_t *load_unicode_string(unsigned id)
{
	static wchar_t str[4096];

	memset(str, 0, sizeof(str));
	LoadStringW(GetModuleHandle(NULL), id, str, lenof(str));

	return str;
}

const char *load_ansi_string(unsigned id)
{
	static char astr[4096];
	wchar_t str[4096];
	unsigned l;

	memset(str, 0, sizeof(str));
	LoadStringW(GetModuleHandle(NULL), id, str, lenof(str));

	l = wcstombs(0, str, 0);
	if (l >= lenof(astr))
		l = lenof(astr) - 1;

	wcstombs(astr, str, l + 1);

	return astr;
}

